#! /usr/bin/env python
# -*- coding: utf-8 -*-


""" This module is made for the joint IFU Rainbow camera astrometry. """

from astropy import units, coordinates, wcs, units
from astropy.io import fits
from astropy.time import Time
import iminuit
import matplotlib.pyplot as mpl
from matplotlib.patches import Polygon, Ellipse
import numpy as np
import pandas
from psfcube import fitter
from pyifu import spectroscopy
from scipy import stats
from scipy.optimize import fmin
from scipy.spatial import distance
import shapely
import warnings


from . import io
from . import sedm


SEDM_SPAXEL_ARCSEC = 0.55
SEDM_LEFT_ROTATION = 2.5 * np.pi/180


# ======================= #
#   FIND POINT SOURCE     #
# ======================= #
MARKER_PROP = {"astrom": dict(marker="x", lw=2, s=80, color="C1", zorder=8),
               "manual": dict(marker="+", lw=2, s=100, color="k", zorder=8),
               "aperture": dict(marker="+", lw=2, s=100, color="k", zorder=8),
               "auto":  dict(marker="o", lw=2, s=200, facecolors="None", edgecolors="C3", zorder=8)
               }


def position_source(cube,
                    centroid=None, centroiderr=None,
                    lbdaranges=[5000, 7000], maxpos=False, warn=True):
    """ How is the source position selected ? """

    if centroiderr is None or centroiderr in ["None"]:
        centroid_err_given = False
        centroids_err = [3, 3]
    else:
        centroid_err_given = True
        centroids_err = np.asarray(centroiderr, dtype="float")

    # Let's start...
    position_type = "unknown"

    # OPTION 1: Use maximum spaxels to derive position
    if maxpos:
        if warn:
            warnings.warn("Centroid guessed based on brightness")
        xcentroid, ycentroid = estimate_default_position(
            cube, lbdaranges=lbdaranges)
        if not centroid_err_given:
            centroids_err = [5, 5]
        position_type = "auto"

    # OPTION 2: Use astrometry of guider images for position, if possible
    elif centroid is None or centroid in ["None"]:
        xcentroid, ycentroid = guess_target_pos(cube.filename)

        if xcentroid is None or ycentroid is None or np.isnan(xcentroid*ycentroid):
            # FAILS... go back to OPTION 1
            if warn:
                warnings.warn("IFU target location based on CCD astrometry failed. "
                              "centroid guessed based on brightness used instead")
            return position_source(cube, lbdaranges=lbdaranges, maxpos=True)

        # Works !
        if warn:
            warnings.warn(
                f"IFU position based on CCD wcs solution used : {xcentroid, ycentroid}")
        position_type = "astrom"

    # OPTION 3: Use input centroid values for position
    else:
        try:
            xcentroid, ycentroid = np.asarray(centroid, dtype="float")
        except:
            raise ValueError(
                "Unable to parse `centroid`. Should be None or a 2d float array `xcentroid, ycentroid = centroids`")
        position_type = "manual"

    warnings.warn(
        f"Position type {position_type}:{xcentroid:.2f}, {ycentroid:.2f}")

    return [xcentroid, ycentroid], centroids_err, position_type


def read_radec(filename=None, header=None):
    """ """
    if header is None:
        header = fits.getheader(filename)

    co = coordinates.SkyCoord(header.get("OBJRA"), header.get("OBJDEC"),
                              frame='icrs', unit=(units.hourangle, units.deg))
    return co.ra.deg, co.dec.deg


def get_wcs_dict(filename=None, radec=None, spxy=None, **kwargs):
    """ """
    if radec is not None and spxy is not None:
        return build_wcs(radec=radec, spxy=spxy, **kwargs)
    elif filename is None:
        raise ValueError(
            "either filename or (radec & spxy) must be given as input")

    astrom = Astrometry(filename)
    return astrom.get_wcs_dict(**kwargs)


def build_wcs(radec, spxy, rotation=2, scale=0.55):
    """ """
    theta = - rotation * np.pi/180
    rot = np.asarray([[np.cos(theta), np.sin(theta)],
                     [-np.sin(theta), np.cos(theta)]])
    scaling = scale * np.asarray([-1/np.cos(radec[1]*np.pi/180), 1])
    offset_radec = np.dot(rot, np.asarray(spxy)/3600)

    ra0, dec0 = radec - scaling*offset_radec

    return {"CTYPE1": 'RA---TAN',
            "CTYPE2": 'DEC--TAN',
            "CRVAL1": ra0, "CRVAL2": dec0,
            "CRPIX1": 1, "CRPIX2": 1,
            "CDELT1": -scale/3600,
            "CDELT2": scale/3600}
# ======================= #
#   High Level            #
# ======================= #


def fit_cube_pos(filename, lbdarange=[6000, 8000], guess="guess", show=False, **kwargs):
    """ """
    sl = spectroscopy.Slice.from_cubefile(filename)
    if guess in ["brightest"]:
        centroid = np.nanmean(np.asarray(
            sl.index_to_xy(sl.get_brightest_spaxels(5))), axis=0)
    else:
        centroid = guess_target_pos(filename)

    fitted_indexes = sl.get_spaxels_within_polygon(
        shapely.geometry.Point(*centroid).buffer(10))
    sl2 = sl.get_subslice(fitted_indexes)
    guess_pos = np.nanmean(np.asarray(
        sl2.index_to_xy(sl2.get_brightest_spaxels(5))), axis=0)

    slf = fitter.fit_slice(sl2, centroids=guess_pos, **kwargs)
    xpos, ypox = [[slf.fitvalues["%scentroid" % ax_],
                   slf.fitvalues["%scentroid.err" % ax_]] for ax_ in ["x", "y"]]
    if show:
        fig = slf.show(zorder=3)
        [ax_.scatter(*centroid, marker="x", color="C1", zorder=6)
         for ax_ in fig.axes[:3]]
        [ax_.scatter(xpos[0], ypox[0], marker=".", color="k", zorder=6)
         for ax_ in fig.axes[:3]]

    del sl2, fitted_indexes
    del sl, centroid, slf
    return xpos, ypox


def filename_to_dateid(filename):
    """ """
    id_ = io.filename_to_id(filename)
    date = io.filename_to_date(filename)
    return date+"_"+id_


def get_allccd_pos(cube_filenames, verbose=True):
    """ """
    # This way, no "too many open files"
    coords = {}
    for fitsfile in cube_filenames:
        dateid = filename_to_dateid(fitsfile)
        coords[dateid] = np.asarray(get_ccd_pos(fitsfile)).tolist()

    return coords


def fit_allcubes_pos(cube_filenames, lbdarange=[6000, 8000], **kwargs):
    """ """
    coords = {}
    for fitfile in cube_filenames:
        dateid = filename_to_dateid(fitfile)
        xpos, ypos = fit_cube_pos(fitfile, lbdarange=lbdarange, **kwargs)
        coords[dateid] = [xpos, ypos]

    return coords
# ======================= #
#   GET LOCATIONS         #
# ======================= #


def guess_target_pos(filename, parameters=None, oldmethod=False):
    """ """
    if oldmethod:
        return oldversion_get_target_pos(filename, parameters=parameters)

    astro = Astrometry(filename)
    if not astro.is_astrom_file_available():
        return None, None
    return astro.get_target_coordinate()


def oldversion_get_target_pos(filename, parameters=None):
    print
    # date
    cube_date = io.filename_to_date(filename, iso=True)

    # = NEWEST VERSION = #
    if Time(cube_date) > Time("2019-04-17"):
        print("INFO: using new astrometric method")
        wcsifu = WCSIFU.from_filename(filename)
        return wcsifu.get_ifu_guess_pos()[0]

    print("WARNING: Soon to be deprecated: old ccd<->ifu structure. *You have nothing to do*")

    if parameters is None:
        parameters = sedm.get_sedm_astrom_param(cube_date)

    # CCD Position
    ccd_xy = get_ccd_pos(filename)
    if Time(cube_date) > Time("2019-02-20") and Time(cube_date) < Time("2019-04-17"):
        # print("TEMPORARY PATCH FOR SHIFTED IFU POS")
        return rainbow_coords_to_ifu(ccd_xy, parameters) + np.asarray([11, 1])
    elif Time(cube_date) > Time("2019-04-22"):
        #print("TEMPORARY PATCH 2 FOR SHIFTED IFU POS")
        return rainbow_coords_to_ifu(ccd_xy, parameters) + np.asarray([-5, -30])
    return rainbow_coords_to_ifu(ccd_xy, parameters)


def get_ccd_pos(filename, radec=None, verbose=True):
    """

    Parameters
    ----------

    """
    astrom_file = io.filename_to_guider(filename)
    if len(astrom_file) == 0:
        if verbose:
            print("No astrom file found for %s" % filename)
        return [np.NaN, np.NaN]

    astrom_file = astrom_file[0]
    with fits.open(filename) as f:
        # Get target coordinates
        header = f[0].header
        if radec is None:
            try:
                radec = coordinates.SkyCoord(header["OBJRA"], header["OBJDEC"],
                                             unit=(units.hourangle, units.deg))
                # Format changed at some point
            except KeyError:
                radec = coordinates.SkyCoord(header["OBRA"], header["OBDEC"],
                                             unit=(units.hourangle, units.deg))
            del header
        elif type(radec) != coordinates.SkyCoord:
            radec = coordinates.SkyCoord(*radec, unit=units.deg)

    # Convert it into ccd pixels
    with fits.open(astrom_file) as f:
        header = f[0].header
        wcs_ = wcs.WCS(header)
        xy = np.asarray(radec.to_pixel(wcs_))
        del header
        del wcs_
        del radec

    return xy


def get_object_ifu_pos(cube, parameters=None):
    """ the expected cube x,y position of the target within the cube """
    print("DEPRECATED WARNING get_object_ifu_pos(cube) -> guess_target_pos(cubefile) ")
    return guess_target_pos(cube.filename)


def get_ccd_coords(cube):
    """ target position in the rainbow camera.
    Remark that this area should not be visible in the rainbow camera as this area
    should be behind the mirror sending the light to the IFU
    """
    print("DEPRECATED WARNING get_ccd_coords(cube) -> get_ccd_pos(cubefile)")
    return get_ccd_pos(cube.filename)


def fit_conversion_matrix(cubes_to_fit, guess=None):
    """ """

    if guess is None:
        guess = sedm.get_sedm_astrom_param(cubes_to_fit)
    else:
        guess = np.asarray(guess)

    list_of_ccd_positions = np.asarray(
        [get_ccd_coords(c_) for c_ in cubes_to_fit])
    list_of_ifu_positions = np.asarray(
        [fit_cube_centroid(c_)[:2] for c_ in cubes_to_fit])

    def to_fit(parameters):
        list_of_ifu_positions_MODEL = rainbow_coords_to_ifu(
            list_of_ccd_positions, parameters)

        return np.sum([distance.euclidean(im_, i_) for im_, i_ in zip(list_of_ifu_positions_MODEL,
                                                                      list_of_ifu_positions)])
    return fmin(to_fit, guess, maxiter=10000)


# ======================= #
#   Other tools           #
# ======================= #
def estimate_default_position(cube, lbdaranges=[5000, 7000]):
    """ """
    sl = cube.get_slice(
        lbda_min=lbdaranges[0], lbda_max=lbdaranges[1], slice_object=True)
    x, y = np.asarray(sl.index_to_xy(sl.indexes)).T  # Slice x and y
    lim_perc = 99.5
    bright_lim = np.nanpercentile(sl.data, lim_perc)
    while np.isnan(bright_lim):
        lim_perc -= 5.
        bright_lim = np.nanpercentile(sl.data, lim_perc)

    argmaxes = np.argwhere(sl.data > bright_lim).flatten()  # brightest points
    return np.nanmedian(x[argmaxes]), np.nanmedian(y[argmaxes])  # centroid


def rainbow_coords_to_ifu(ccd_coords, parameters):
    """ """
    centroid_ccd = np.asarray([parameters[4], parameters[5]])
    matrix = np.asarray([[parameters[0], parameters[1]],
                         [parameters[2], parameters[3]]])

    return np.dot(matrix, (ccd_coords-centroid_ccd).T).T


def fit_cube_centroid(cube_, lbdamin=6000, lbdamax=7000):
    """ Use `fit_slice` function from psfcube to estimate the PSF centroid
    of a point-source containing in the cube.
    The fit will be made on the metaslice within the boundaries `lbdamin` `lbdamax`

    Parameters
    ----------
    cube_: [pyifu Cube]
        Cube containing the point source

    lbdamin, lbdamax: [floats] -optional-
        lower and upper wavelength boundaries defining the metaslice
        [in Angstrom]

    Returns
    -------
    list: [x,y, dx, dy] # (centroids and errors)
    """
    sl_r = cube_.get_slice(
        lbda_min=lbdamin, lbda_max=lbdamax, slice_object=True)
    centroid = estimate_default_position(cube_)
    slfit = fitter.fit_slice(sl_r, centroids=centroid, centroids_err=[4, 4])
    return [slfit.fitvalues[k] for k in ["xcentroid", "ycentroid", "xcentroid.err", "ycentroid.err"]]


def get_standard_rainbow_coordinates(rainbow_wcs, stdname):
    """ """
    try:
        import pycalspec
    except ImportError:
        raise ImportError("Cannot find pycalspec module. Please install it first")
    rhms = coordinates.SkyCoord(*pycalspec.std_radec(stdname),
                                unit=(units.hourangle, units.deg))
    return np.asarray(rhms.icrs.to_pixel(rainbow_wcs))

# ======================== #
#                          #
#    Class                 #
#                          #
# ======================== #


class Astrometry():
    # Manages connections between RA-Dec | IFU | Guider coordinates

    def __init__(self, filename=None, isodate=None):
        """ """
        if filename is not None:
            self.filename = filename
            self._load_parameters_(io.filename_to_date(filename, iso=True))
        if isodate is not None:
            self._load_parameters_(isodate)

    # ================ #
    #  Methods         #
    # ================ #
    def _load_parameters_(self, date):
        """ """
        self.date = date
        scale = 0.55
        rotation = 1
        position = np.NaN, np.NaN

        if Time(date) >= Time("2020-10-04"):
            position = 1039.5, 968.5
            scale = 0.55
            rotation = 1
        elif Time(date) >= Time("2020-10-03"):
            position = 1019.0, 945.0
            scale = 0.55
            rotation = 1

        elif Time(date) >= Time("2020-08-28"):
            position = 1027.5, 950.5
            scale = 0.55
            rotation = 1

        elif Time(date) >= Time("2020-07-10"):
            position = 1023.5, 958.5
            scale = 0.55
            rotation = 1

        elif Time(date) > Time("2019-09-06"):
            position = 1026.5, 976
            scale = 0.55
            rotation = 1

        elif Time(date) > Time("2019-06-17"):
            position = 1026.5, 975.3
            scale = 0.55
            rotation = 1

        elif Time(date) > Time("2019-05-24"):
            position = 1029.5, 975.3
            scale = 0.55
            rotation = 1

        elif Time(date) > Time("2019-05-14"):
            position = 1026.5, 975.3
            scale = 0.55
            rotation = 1

        elif Time(date) > Time("2019-05-13"):
            print("WARNING: ROTATION.")
            print("CONTACT TO MICKAEL IF YOU NEED DATA FOR THIS DATE.")
            position = 1033.5, 975.3
            scale = 0.55
            rotation = 1

        elif Time(date) > Time("2019-04-23"):
            position = 1027, 978
            scale = 0.55
            rotation = 2

        elif Time(date) > Time("2019-04-18"):
            position = 1022, 1025
            scale = 0.55
            rotation = 2

        elif Time(date) > Time("2019-02-28"):
            position = 1017, 1015
            scale = 0.55
            rotation = 2

        elif Time(date) > Time("2019-01-05"):
            position = 1033, 1017
            scale = 0.55
            rotation = 2

        elif Time(date) > Time("2018-10-17"):
            position = 1016.5, 1015.5
            scale = 0.55
            rotation = 2

        elif Time(date) > Time("2018-09-26"):
            position = 1019.5, 1014.5
            scale = 0.55
            rotation = 2

        else:
            position = 1027.5, 1016.0
            scale = 0.55
            rotation = 2

        self.parameters = {"position": position,
                           "scale": scale,
                           "rotation": rotation}

    # -------- #
    #  SETTER  #
    # -------- #
    def set_guider_offset(self, offsetx, offsety):
        """ """
        self._guider_offset = np.asarray([offsetx, offsety])

    def set_ifu_offset(self, spaxelx, spaxely):
        """ """
        self._ifu_offset = np.asarray([spaxelx, spaxely])

    # -------- #
    #  LOADER  #
    # -------- #
    def load_cube(self):
        """ """
        self.cube = sedm.get_sedmcube(self.filename)
        self.cube.load_adr()

    def load_slice(self, lbdarange=[5000, 6000]):
        """ """
        self.slice = spectroscopy.Slice.from_cubefile(
            self.filename, lbdarange=lbdarange)

    def load_guider(self, load_gaiacat=True):
        """ """
        try:
            from astrobject import photometry
        except ImportError:
            raise ImportError("You need astrobject to load the guider")
        self._guider = photometry.get_image(self.astromfile, background=0)
        if load_gaiacat:
            self._guider.download_catalogue("gaia")

    # -------- #
    #  GETTER  #
    # -------- #
    def get_target_coordinate(self, where="ifu"):
        """ """
        return self.radec_to(where, *self.target_radec)

    def get_wcs_dict(self, spaxel_size_arcsec=0.55):
        """ """
        spra, spdec = 0, 0
        ra, dec = self.ifu_to_radec(spra, spdec)
        return {"CTYPE1": 'RA---TAN',
                "CTYPE2": 'DEC--TAN',
                "CRVAL1": ra, "CRVAL2": dec,
                "CRPIX1": spra+1, "CRPIX2": spdec+1,
                "CDELT1": -spaxel_size_arcsec/3600,
                "CDELT2": spaxel_size_arcsec/3600}
    # -------- #
    # CONVERT  #
    # -------- #

    def ifu_to(self, where, spaxelx, spaxely):
        """ """
        return self.radec_to(where, *self.ifu_to_radec(spaxelx, spaxely))

    def guider_to(self, where, x, y):
        """ """
        if where in ["guider", "guiderxy", "astrom", "astromxy"]:
            return np.asarray([x, y])
        if where in ["coords", "radec", "world", "sky"]:
            return self.wcs_guider.all_pix2world(np.asarray([np.atleast_1d(x), np.atleast_1d(y)]).T, 0)[0]

        return self.radec_to(where, *self.guider_to("radec", x, y))

    def radec_to(self, where, ra, dec):
        """

        Parameters
        ----------
        where: [string]
            The destination, it could be:
            - guider/astrom: guider image ccd (pixels)
            - ifu: IFU (spaxels)

        ra, dec: [float, float or list of]
            Input Coordinates

        Returns
        -------
        same format of ra, dec
        """
        # CCD Positions (reference image or guider)
        if where in ["guider", "guiderxy", "astrom", "astromxy"]:
            return self.wcs_guider.all_world2pix(np.asarray([np.atleast_1d(ra), np.atleast_1d(dec)]).T, 0)[0]

        # IFU Position
        if where in ["ifu"]:
            return self.radec_to_ifu(ra, dec)

        raise ValueError(
            "Cannot parse the input `where` (%s) could be: 'ref'/'guider'/'ifu'" % where)

    def radec_to_ifu(self, ra, dec):
        """ """
        scaling = np.asarray([-self.parameters["scale"]/np.cos(self.ifucentroid_radec[1]*np.pi/180),
                              self.parameters["scale"]])

        offset_spaxels = (np.asarray(
            [ra, dec]).T-self.ifucentroid_radec)*units.deg.to("arcsec") / scaling
        theta = self.parameters["rotation"] * np.pi/180
        rot = np.asarray([[np.cos(theta), np.sin(theta)],
                         [-np.sin(theta), np.cos(theta)]])
        return (np.dot(rot, offset_spaxels.T).T + self.ifu_offset).T

    def ifu_to_radec(self, spaxelx, spaxely):
        """ """
        scaling = np.asarray([-self.parameters["scale"]/np.cos(self.ifucentroid_radec[1]*np.pi/180),
                             self.parameters["scale"]])

        theta = - self.parameters["rotation"] * np.pi/180
        rot = np.asarray([[np.cos(theta), np.sin(theta)],
                         [-np.sin(theta), np.cos(theta)]])
        offset_spaxels = np.dot(
            rot, (np.asarray([spaxelx, spaxely]).T - self.ifu_offset).T)
        return ((offset_spaxels.T * scaling) / units.deg.to("arcsec") + self.ifucentroid_radec).T

    def is_astrom_file_available(self):
        """ """
        # Already has an astromfile loaded.
        if hasattr(self, "_astromfile") and self._astromfile is not None:
            return True

        fileastrom = io.filename_to_guider(self.filename)
        if fileastrom is None or len(fileastrom) == 0:
            return False
        return True

    # -------- #
    # PLOTTER  #
    # -------- #

    def show_guider_wcs(self, **kwargs):
        """ """
        return self.guider.show(**{**dict(show_catalogue=True), **kwargs})

    # ================ #
    #  Properties      #
    # ================ #

    @property
    def astromfile(self):
        """ """
        if not hasattr(self, "_astromfile"):
            try:
                self._astromfile = io.filename_to_guider(self.filename)[0]
            except:
                raise IOError(f"No astrometric file for {self.filename}")

        return self._astromfile

    @property
    def wcs_guider(self):
        """ """
        if not hasattr(self, "_wcsguider"):
            header = fits.getheader(self.astromfile)
            self._wcsguider = wcs.WCS(header)
            del header
        return self._wcsguider

    @property
    def guider(self):
        """ """
        if not hasattr(self, "_guider"):
            self.load_guider()
        return self._guider

    @property
    def target_radec(self):
        """ RA and Dec of the target (as stored in the hearder)"""
        if not hasattr(self, "_target_radec"):
            header = fits.getheader(self.filename)
            _scoord = coordinates.SkyCoord(header.get("OBJRA"), header.get(
                "OBJDEC"), unit=(units.hour, units.deg))
            del header
            self._target_radec = _scoord.ra.deg, _scoord.dec.deg

        return self._target_radec

    @property
    def guider_offset(self):
        """ Offset of the guider ccd pixel corresponding to the IFU centroid (spaxel 0,0) """
        if not hasattr(self, "_guider_offset"):
            self.set_guider_offset(0, 0)
        return self._guider_offset

    @property
    def ifu_offset(self):
        """ Offset of the ifu centroid (spaxel 0,0 -> ifu_offset) """
        if not hasattr(self, "_ifu_offset"):
            self.set_ifu_offset(0, 0)
        return self._ifu_offset

    @property
    def ifucentroid_radec(self):
        """ """
        return self.wcs_guider.all_pix2world([self.parameters["position"] + self.guider_offset], 0)[0]


class IFUReference(Astrometry):

    def __init__(self, filename=None, isodate=None, load_refimage=True, **kwargs):
        """ """
        _ = super().__init__(filename=filename, isodate=isodate)

        self.load_slice([7000, 7500])

        if load_refimage:
            try:
                self.load_refimage()
                self.reload_refimage_sep(0.1)
            except:
                warnings.warn("Cannot load the reference image")

    @classmethod
    def from_astrometry(cls, astrometry, load_refimage=True):
        """ """
        this = cls(astrometry.filename, load_refimage=False)
        if hasattr(astrometry, "_guider"):
            this._astromfile = astrometry._astromfile
            this._guider = astrometry._guider
        if hasattr(astrometry, "slice"):
            this.slice = astrometry.slice
        if hasattr(astrometry, "cube"):
            this.cube = astrometry.cube

        # offset if any
        this.set_ifu_offset(*astrometry.ifu_offset)
        this.set_guider_offset(*astrometry.guider_offset)
        if load_refimage:
            this.load_refimage()

        return this

    # ================ #
    #  Methods         #
    # ================ #
    # ------- #
    #  SETTER #
    # ------- #
    # ------- #
    # LOADER  #
    # ------- #

    def load_refimage(self):
        """ """
        try:
            from photoifu import photoref
        except:
            raise ImportError("Cannot import photoifu, please install it.")
        self.photoref = photoref.PhotoReference.from_coords(self.target_radec)
        self._sources_xy = self.refimage.sepobjects.get(["x", "y"]).T

    def reload_refimage_sep(self, thresh=1, **kwargs):
        """ """
        self.refimage.sep_extract(thresh)
        self._sources_xy = self.refimage.sepobjects.get(["x", "y"]).T

    # ------- #
    # GETTER  #
    # ------- #
    def get_iso_contours(self, where="ref", targetmag=None, isomag=None):
        """ """
        if targetmag is not None:
            self.photoref.build_pointsource_image(targetmag)
        if isomag is None:
            isomag = np.linspace(20, 26, 13)

        contours_ref = self.photoref.get_iso_contours(isomag, asdict=True)
        if where in ["reference", "ref", "refimage", "refccd", "refxy"]:
            return contours_ref
        return {iso: [self.reference_to(where, *cc.T).T for cc in conts] for iso, conts in contours_ref.items()}

    def get_refimage_sources_coordinates(self, where="ref", inifu=False):
        """ """
        xpos, ypos = self._sources_xy
        if where in ["ref"]:
            x_, y_ = np.asarray([xpos, ypos])
        else:
            ra, dec = self.refimage.pixel_to_coords(xpos, ypos).T
            if where in ["world", "coords", "radec"]:
                x_, y_ = np.asarray([ra, dec])
            else:
                x_, y_ = np.asarray(self.radec_to(where, ra, dec))

        if inifu:
            flagin = self.slice.contains(
                *self.get_refimage_sources_coordinates("ifu", inifu=False))
            return np.asarray([x_[flagin], y_[flagin]])

        return np.asarray([x_, y_])

    def get_fakeslice(self, mag, seeing=None, refseeing=1, slicearcsec=0.55):
        """ """
        self.photoref.build_pointsource_image(
            mag, refseeing/self.photoref.refimage.pixel_size_arcsec.value)
        if seeing is not None:
            addseeing = np.sqrt(seeing**2 - refseeing**2)
        else:
            addseeing = None
        return self.photoref.project_to_slice(self.slice, slicearcsec, self.get_target_coordinate("ifu"),
                                              gaussianconvol=addseeing / self.photoref.refimage.pixel_size_arcsec.value)

    def get_slicemodeller(self, mag, seeing=1.1, **kwargs):
        """ """
        try:
            from photoifu import slice
        except:
            raise ImportError("Cannot import photoifu, please install it")
        smodel = slice.SliceModeller(self.slice, self.photoref)
        smodel.set_target(self.get_target_coordinate(), mag)
        smodel.load_modelslice(seeing=seeing, **kwargs)
        smodel.load_gridslice(True)
        smodel.load_gridslice(False)
        return smodel

    # ------- #
    # PLOTTER #
    # ------- #
    def show(self, show_title=True, ifuvmax="99", ifuvmin="5"):
        """ """
        fig = mpl.figure(figsize=[6, 3])
        axphoto = fig.add_axes([0.08, 0.135, 0.4, 0.7])
        axifu = fig.add_axes([0.55, 0.135, 0.35, 0.7])

        self.show_image(ax=axphoto)
        self.show_slice(ax=axifu, show_colorbar=False,
                        vmax=ifuvmax, vmin=ifuvmin)

        xy_edges = self.radec_to(
            "ref", *self.ifu_to_radec(*np.asarray(self.slice.convexhull_vertices)))
        axphoto.add_patch(Polygon(xy_edges, facecolor="None",
                          edgecolor=mpl.cm.viridis(0), linewidth=2))
        axifu.set_yticks([])
        axifu.set_xticks([])

        if show_title:
            title = "%s" % self.slice.header["OBJECT"].split()[0]+" | " + \
                    "RA: %.4f Dec: %.4f" % (self.target_radec) + " | " + \
                    "airmass: %.2f" % self.slice.header["AIRMASS"] + "\n" + \
                    "AZ %.2f" % self.slice.header["TEL_AZ"] + " | " + \
                    "EL %.1f" % self.slice.header["TEL_EL"] + " | " + \
                    "parangle %.1f" % self.slice.header["TEL_PA"]

            _ = fig.text(0.5, 0.975, title, va="top", ha="center", color="0.7")

        return fig

    def _show_sources_(self, ax, which, **kwargs):
        """ """
        flagin = self.slice.contains(
            *self.get_refimage_sources_coordinates("ifu"))
        sources = self.get_refimage_sources_coordinates(which)
        ax.plot(*sources.T[flagin].T, color="k", **self._marker_prop)
        ax.plot(*sources.T[~flagin].T, color="k",
                alpha=0.3, **self._marker_prop)
        ax.plot(*self.get_target_coordinate(which),
                color="C1", **self._marker_prop)

    def show_image(self, ax=None, **kwargs):
        """ """
        _ = self.refimage.show(ax=ax, show_target=False, **kwargs)
        if ax is None:
            ax = _["ax"]

        self._show_sources_(ax, "ref", **kwargs)

    def show_slice(self, ax=None, **kwargs):
        """ """
        fig = self.slice.show(ax=ax, **{**dict(zorder=4), **kwargs})
        if ax is None:
            ax = fig.axes[0]

        self._show_sources_(ax, "ifu", **kwargs)

    def show_segmap(self, **kwargs):
        """ """
        return self.photoref.show(**kwargs)

    @property
    def _marker_prop(self):
        """ """
        return dict(ls="None", ms=8, mew=1.5, marker="x", scalex=False, scaley=False, zorder=5)

    # ------- #
    # CONVERT #
    # ------- #
    def reference_to(self, where, x, y):
        """ Generic convertion tool to go from pixels in the  reference images
        to anywhere.

        Parameters
        ----------
        where: [string]

            The destination, it could be:
            - ref/refimage/refxy: reference image ccd (pixels)
            - guider/astrom: guider image ccd (pixels)
            - coords/world/radec: RA and Dec
            - ifu: IFU (spaxels)

        """
        if where in ["reference", "ref", "refimage", "refccd", "refxy"]:
            return np.asarray([x, y])

        ra, dec = self.refimage.pixel_to_coords(x, y).T
        if where in ["coords", "radec", "world", "sky"]:
            return ra, dec

        return self.radec_to(where, ra, dec)

    def radec_to(self, where, ra, dec):
        """

        Parameters
        ----------
        where: [string]

            The destination, it could be:
            - ref/refimage/refxy: reference image ccd (pixels)
            - guider/astrom: guider image ccd (pixels)
            - ifu: IFU (spaxels)

        ra, dec: [float, float or list of]
            Input Coordinates

        Returns
        -------
        same format of ra, dec
        """
        # CCD Positions (reference image or guider)
        if where in ["reference", "ref", "refimage", "refccd", "refxy"]:
            return self.refimage.coords_to_pixel(ra, dec)

        return super().radec_to(where, ra, dec)

    def reference_to_ifu(self, x, y):
        """ """
        return self.reference_to("ifu", x, y)

    def ifu_to_reference(self, spaxelx, spaxely):
        """ """
        return self.ifu_to("ref", spaxelx, spaxely)

    # ================ #
    #  Properties      #
    # ================ #
    @property
    def refimage(self):
        """ Reference image """
        return self.photoref.refimage

    @property
    def refcatdata(self):
        """ """
        return self.photoref.catdata


class SliceAligner():
    """ """
    FREEPARAMETERS = ["theta", "offsetx", "offsety"]

    def __init__(self, slice_, source_position):
        """ """
        print("*** SliceAligner is DEPRECATED ***")
        self.slice = slice_
        self.set_sourcepos(*source_position)

    # =========== #
    #  Methods    #
    # =========== #
    def _data_to_slice_(self, data):
        """ """
        return spectroscopy.get_slice(data,
                                      self.slice.xy.T,
                                      spaxel_vertices=self.slice.spaxel_vertices,
                                      indexes=self.slice.indexes,
                                      lbda=self.slice.lbda)

    def load_slicegrid(self, size=None):
        """ """
        self._grid = self.get_slicegrid(size=size)

    def load_slicesources(self, usepeak=False, **kwargs):
        """ """

        if not hasattr(self, "dfsep"):
            self.extract_slicesource(**kwargs)

        keys = ["x", "y"] if not usepeak else ["xpeak", "ypeak"]
        self._slicesources = self.dfsep[keys].values.T
        self._slicesources_sky = coordinates.SkyCoord(
            *self._slicesources, unit=units.arcsec)
        self._ellipses = {"x": self.dfsep[keys[0]].values, "y": self.dfsep[keys[1]].values,
                          "a": self.dfsep["a"].values, "b": self.dfsep["b"].values,
                          "theta": self.dfsep["theta"].values}

    def extract_slicesource(self, thresh=1e-4, **kwargs):
        """
        Properties
        ----------
        thresh:

        kwargs goes to sep.extract(thresh=thresh, **kwargs)
            # Default (by sep.)
            mask=None, minarea=5,
            filter_kernel=default_kernel, filter_type='matched',
            deblend_nthresh=32, deblend_cont=0.005 (customed), clean=True,
            clean_param=1.0, segmentation_map=False
            maskthresh : float, optional

            # Customisation
            - deblend_cont=1e-4

            # documentation
            - minarea : int, optional
                Minimum number of pixels required for an object. Default is 5.
            - filter_kernel : `~numpy.ndarray` or None, optional
                Filter kernel used for on-the-fly filtering (used to
                enhance detection). Default is a 3x3 array:
                [[1,2,1], [2,4,2], [1,2,1]]. Set to ``None`` to skip
                convolution.
            - filter_type : {'matched', 'conv'}, optional
                Filter treatment. This affects filtering behavior when a noise
                array is supplied. ``'matched'`` (default) accounts for
                pixel-to-pixel noise in the filter kernel. ``'conv'`` is
                simple convolution of the data array, ignoring pixel-to-pixel
                noise across the kernel.  ``'matched'`` should yield better
                detection of faint sources in areas of rapidly varying noise
                (such as found in coadded images made from semi-overlapping
                exposures).  The two options are equivalent when noise is
                constant.
            - deblend_nthresh : int, optional
                Number of thresholds used for object deblending. Default is 32.
            - deblend_cont : float, optional
                Minimum contrast ratio used for object deblending.
                To entirely disable deblending, set to 1.0.
            - clean : bool, optional
                Perform cleaning? Default is True.
            - clean_param : float, optional
                Cleaning parameter (see SExtractor manual). Default is 1.0.
            - segmentation_map : bool, optional
                If True, also return a "segmentation map" giving the member
                pixels of each object. Default is False.
        """
        try:
            import sep
        except ImportError:
            raise ImportError("Cannot import sep. Please install sep.")
        flagin = self.slice.contains(*self.grid.geodataframe[["x", "y"]].values.T,
                                     buffer=-2)

        data = self.grid.geodataframe["data"].values.copy()
        # set nans to things outside the slice
        data[~flagin] = np.NaN
        # Normalize everything
        data = (data - np.percentile(data[data == data], 1)) / (np.percentile(
            data[data == data], 99) - np.percentile(data[data == data], 1))
        # reshape
        self.imgdata = data.reshape(self.gridsize*2, self.gridsize*2)
        # run sep.extract and convert results in pandas
        extract_prop = {**{"deblend_cont": 1e-4}, **kwargs}
        self.dfsep = pandas.DataFrame(sep.extract(
            self.imgdata, thresh, **extract_prop))
        # Shift the grid to be aligned with slice coordinates
        keys = ["xmin", "xmax", "ymin", "ymax", "x", "y", "xpeak", "ypeak"]
        self.dfsep[keys] = self.dfsep[keys].apply(lambda x: x-self.gridsize)
        return self.dfsep

    # ------- #
    # FITTER  #
    # ------- #
    # - Distance
    def get_source_matching(self):
        """ """
        return self.slicesources_skycoord.match_to_catalog_sky(
            coordinates.SkyCoord(*self.source_effposition, unit=units.arcsec)
        )

    # - Initialisation
    def get_initial_guess(self, asarray=False, **kwargs):
        """ """
        dict_ = {**{t: 0 for t in self.freeparameters}, **
                 {"theta": 0, "offsetx": 0, "offsety": 0}, **kwargs}
        if not asarray:
            return dict_
        return [dict_[t] for t in self.freeparameters]

    def get_boundaries(self, asarray=False, **kwargs):
        """ """
        dict_ = {**{t: [None, None] for t in self.freeparameters},
                 **{"theta": [-np.pi, np.pi], "offsetx": [-5, 5], "offsety": [-3, 2]}, **kwargs}
        if not asarray:
            return dict_
        return [dict_[t] for t in self.freeparameters]

    def get_fiterror(self, asarray=False, **kwargs):
        """ """
        dict_ = {**{t: 0.1 for t in self.freeparameters}, **kwargs}
        if not asarray:
            return dict_
        return [dict_[t] for t in self.freeparameters]

    def get_fixed(self, asarray=False, **kwargs):
        """ """
        dict_ = {**{t: False for t in self.freeparameters},
                 **{"theta": True}, **kwargs}
        if not asarray:
            return dict_
        return [dict_[t] for t in self.freeparameters]

    # - Prior
    def get_logpriors(self, parameters=None):
        """
        product of all the single parameter priors
        """
        if parameters is not None:
            self.set_parameters(parameters)

        # List of priors
        return 0

    # - Likelihood
    def get_loglikelihood(self, parameters=None):
        """
        returns likelihood (L), such that
        chi2 = -2*np.sum(np.log(L))
        """
        if parameters is not None:
            self.set_parameters(parameters)

        distance = self.get_source_matching()[1].arcsec
        loglikelihood = stats.norm.logpdf(distance,
                                          loc=0,
                                          scale=1,
                                          )

        return loglikelihood

    def get_chi2(self, parameters=None):
        """ """
        if parameters is not None:
            self.set_parameters(parameters)

        chi2 = -2 * np.sum(self.get_loglikelihood())
        return chi2

    def get_priored_chi2(self, parameters=None):
        """ """
        if parameters is not None:
            self.set_parameters(parameters)

        return self.get_chi2() - 2*np.sum(self.get_logpriors())

    def fit(self, use_prior=True, fixed_prop={}, guess_prop={}, boundaries_prop={}, error_prop={}):
        """
        """
        # 1. Setup iminuit
        # 2. call minuit
        # 3. return and store value

        # need to be tested and fixed!!!
        #limit_lbda0_ = (6563, self.parameters["lbda0"]+2*self.parameters["width"])
        guesses = self.get_initial_guess(**guess_prop)
        boundaries = self.get_boundaries(**boundaries_prop)
        fiterr = self.get_fiterror(**error_prop)
        fixed = self.get_fixed(**fixed_prop)

        self._function_to_minimize = self.get_priored_chi2 if use_prior else self.get_chi2
        self._minuit = iminuit.Minuit(self._function_to_minimize, **guesses)
        self._minuit.errordef = 1
        for limit in boundaries:
            self._minuit.limits[limit] = boundaries[limit]
        for error in fiterr:
            self._minuit.errors[error] = fiterr[error]
        for fix in fixed:
            if fixed[fix] is True:
                self._minuit.fixed[fix] = True
        
        # Run the minimisation
        self._minuit.migrad()

        # dict(param:bestvalue, param.err:errorvalue)
        self.fitvalues = self._minuit.values
        return self.fitvalues

    # ------- #
    # SETTER  #
    # ------- #

    def set_sourcepos(self, sourcex, sourcey):
        """ """
        self.sourcepos = np.asarray([sourcex, sourcey])

    def set_parameters(self, parameters):
        """ """
        self.parameters = {k: v for k, v in zip(
            self.freeparameters, parameters)}

    def set_initial_guess(self, **kwargs):
        """ """
        guesses = self.get_initial_guess(**kwargs)
        self.set_parameters([guesses[k] for k in self.freeparameters])

    # ------- #
    # GETTER  #
    # ------- #
    def get_effective_position(self, ifux, ifuy):
        """ """
        return np.asarray([ifux+self.parameters["offsetx"], ifuy+self.parameters["offsety"]])

    def get_slicegrid(self, size=None):
        """ """
        try:
            from pixelproject import grid #TOFIX: not released on pypi?
        except ImportError:
            raise ImportError("Cannot import pixelproject. Please install it first.")
        if size is None:
            self.gridsize = int(
                np.max(np.abs(self.slice.convexhull_vertices)))+1
        else:
            self.gridsize = size

        pixels = np.mgrid[0:self.gridsize*2, 0:self.gridsize*2]-self.gridsize
        pixels_flat = np.concatenate(pixels.T, axis=0)

        gg = grid.Grid(pixels_flat)
        flagnan = np.asarray(np.sum(np.isnan(self.xy), axis=0), dtype="bool")
        sg = grid.Grid(self.xy.T[~flagnan], self.slice.spaxel_vertices)
        sg.add_data(self.slice.data[~flagnan], "data")
        if self.slice.has_variance():
            sg.add_data(self.slice.variance[~flagnan], "variance")
        return sg.project_to(gg)

    # -------- #
    # PLOTTER  #
    # -------- #
    def show_extraction(self, ax=None, ellipses=True, sourcecolor="C1", **kwargs):
        """ """
        if ax is None:
            fig = mpl.figure(figsize=[6, 5])
            ax = fig.add_axes([0.15, 0.15, 0.75, 0.75])
        else:
            fig = ax.figure

        ax.imshow(self.imgdata, origin="lower", extent=[-self.gridsize, self.gridsize,
                                                        -self.gridsize, self.gridsize], **kwargs)

        ax.scatter(*self.slicesources, marker="x", color=sourcecolor)
        if ellipses:
            for i, pf in pandas.DataFrame(self._ellipses).iterrows():
                ax.add_patch(Ellipse([pf.x, pf.y], pf.a*2, pf.b*2,
                                     angle=pf.theta*180/np.pi, facecolor="None",
                                     edgecolor=sourcecolor)
                             )

        return fig

    def show(self):
        """ """
        fig = self.slice.show(vmax="98", zorder=3)
        ax = fig.axes[0]
        ax.scatter(*self.slicesources, color="C1", marker="x", zorder=4,
                   label="detected")
        ax.scatter(*self.sourcepos, color="k", marker=".", zorder=5,
                   label="expected")
        ax.scatter(*self.source_effposition, color="k", marker="+", zorder=5,
                   label="fitted")

        ax.text(0, 1.02, "Sources:", va="bottom",
                ha="left", transform=ax.transAxes)
        ax.legend(loc=[0.15, 1.01], ncol=3, frameon=False)
        return fig

    # =========== #
    #  Properties #
    # =========== #
    @property
    def nsources(self):
        """ """
        return len(self.sourcepos.T)

    @property
    def source_effposition(self):
        """ """
        return self.get_effective_position(*self.sourcepos)

    @property
    def freeparameters(self):
        """ """
        if not hasattr(self, "_freeparameters"):
            self._freeparameters = self.FREEPARAMETERS.copy()
        return self._freeparameters

    # - from Slice
    @property
    def normalization(self):
        """ """
        if not hasattr(self, "_normalization"):
            self._normalization = np.nanmean(self.slice.data)
        return self._normalization

    @property
    def data(self):
        """ data of the input slice"""
        if not hasattr(self, "_data"):
            self._data = self.slice.data/self.normalization
        return self._data

    @property
    def variance(self):
        """ variance of the slice's data"""
        if not hasattr(self, "_variance"):
            self._variance = self.slice.variance/self.normalization**2
        return self._variance

    @property
    def xy(self):
        """ position of the slice's spaxels """
        return self.slice.xy

    # - Slice Source
    @property
    def slicesources(self):
        """ """
        if not hasattr(self, "_slicesource"):
            self.load_slicesources()
        return self._slicesources

    @property
    def slicesources_skycoord(self):
        """ """
        if not hasattr(self, "_slicesource_sky"):
            self.load_slicesources()
        return self._slicesources_sky

    # - Grid
    @property
    def grid(self):
        """ """
        if not hasattr(self, "_grid"):
            self.load_slicegrid()
        return self._grid


class AstrometryFitter():
    def __init__(self, cube_filenames):
        """ """
        print("*** AstrometryFitter will soon be DEPRECATED ***")
        self.input_files = {filename_to_dateid(f): f for f in cube_filenames}
        self.load_ccd_pos(cube_filenames)
        self.load_cube_pos(cube_filenames)

    def load_ccd_pos(self, cube_filenames=None, validation=False):
        """ """
        coords = get_allccd_pos(cube_filenames)
        attrname = "ccd_xy%s" % ("_validation" if validation else "")
        setattr(self, attrname, coords)

    def load_cube_pos(self, cube_filenames=None, validation=False):
        """ """
        coords = fit_allcubes_pos(cube_filenames)
        attrname = "cube_xy%s" % ("_validation" if validation else "")
        setattr(self, attrname, coords)

    def _load_dataframe_(self):
        """ """
        dict_out = {}
        for id_ in [i for i in self.cube_xy.keys() if i in self.ccd_xy.keys()]:
            dict_out[id_] = {}
            dict_out[id_]["spaxels_x"], dict_out[id_]["spaxels_y"] = np.asarray(
                self.cube_xy[id_])[:, 0]
            dict_out[id_]["ccd_x"], dict_out[id_]["ccd_y"] = np.asarray(
                self.ccd_xy[id_])

        self._coordinates = pandas.DataFrame(dict_out).T
        delta_x, delta_y = (self.transformation(self.coordinates[["ccd_x", "ccd_y"]]) -
                            self.coordinates[["spaxels_x", "spaxels_y"]].values).T
        self.coordinates["delta_x"] = delta_x
        self.coordinates["delta_y"] = delta_y
        self.coordinates["delta"] = np.sqrt(delta_x**2 + delta_y**2)

    def load_transform(self, ttype="affine", **kwargs):
        """

        Parameters
        -----------
        ttype : [string] -optional-
            Type of transform.
            could be:
            'euclidean', similarity', 'affine', 'piecewise-affine', 'projective', 'polynomial'

        **kwargs
            Function parameters (src, dst, n, angle)::

            NAME / TTYPE        FUNCTION PARAMETERS
            'euclidean'         `src, `dst`
            'similarity'        `src, `dst`
            'affine'            `src, `dst`
            'piecewise-affine'  `src, `dst`
            'projective'        `src, `dst`
            'polynomial'        `src, `dst`, `order` (polynomial order,
                                                      default order is 2)

        """
        try:
            from skimage import transform
        except ImportError:
            raise ImportError("Cannot import skimage. Please install scikit-image first.")
        self.flagout = np.asarray(
            np.nansum(np.isnan(self.coordinates).values, axis=1), dtype="bool")
        dd = self.coordinates[~self.flagout]
        tform = transform.estimate_transform(ttype,
                                             dd[["ccd_x", "ccd_y"]].values,
                                             dd[["spaxels_x", "spaxels_y"]].values,
                                             **kwargs)

        self._transform = tform

    def get_residual_distance(self):
        """ """
        dd = self.coordinates[~self.flagout]
        spaxel_pos = dd[["spaxels_x", "spaxels_y"]].values
        ccd_pos = dd[["ccd_x", "ccd_y"]].values
        spaxel_pos-self.transformation(ccd_pos)
        return spaxel_pos

    def show(self):
        """ """
        fig = mpl.figure(figsize=[9, 3])
        axccd = fig.add_axes([0.08, 0.1, 0.27, 0.8])
        axcube = fig.add_axes([0.4, 0.1, 0.27, 0.8])
        axmatch = fig.add_axes([0.72, 0.1, 0.27, 0.8])

        dd = self.coordinates[~self.flagout]
        spaxel_pos = dd[["spaxels_x", "spaxels_y"]].values
        ccd_pos = dd[["ccd_x", "ccd_y"]].values
        npoints = len(dd)
        colors = mpl.cm.viridis(np.arange(len(dd))/len(dd))
        axccd.scatter(*ccd_pos.T, c=colors)
        axcube.scatter(*spaxel_pos.T, c=colors)

        axmatch.scatter(*(spaxel_pos-self.transformation(ccd_pos)).T,
                        c=colors, marker="x", s=100)
        # // Titles
        axccd.set_title("RC-Expectation")
        axcube.set_title("IFU Measurement")
        axmatch.set_title(r"$\Delta$  IFU pos-prediction")

    # =============== #
    #   Properties    #
    # =============== #
    @property
    def ntargets(self):
        """ """
        return len(self.coordinates)

    @property
    def coordinates(self):
        """ """
        if not hasattr(self, "_coordinates") or self._coordinates is None:
            if hasattr(self, "ccd_xy") and hasattr(self, "cube_xy"):
                self._load_dataframe_()
            else:
                self._coordinates = None
        return self._coordinates

    @property
    def transformation(self):
        """ """
        if not hasattr(self, "_transform"):
            self.load_transform()
        return self._transform


class WCSIFU():
    """ """

    def __init__(self, date=None):
        """ """
        print("*** WCSIFU will soon be DEPRECATED ***")
        if date is not None:
            self.load_transform(date)

    # ============== #
    #   Get          #
    # ============== #
    def coords_to_ccdxy(self, filename=None):
        """ """
        if filename is None:
            if not hasattr(self, "filename"):
                raise ValueError("no filename loaded, you must provide one.")

            filename = self.filename

    # ============== #
    #   Loader       #
    # ============== #
    @classmethod
    def from_filename(cls, filename):
        """ """
        date = io.filename_to_date(filename, iso=True)
        this = cls.from_date(date)
        this.filename = filename
        return this

    @classmethod
    def from_date(cls, date):
        """ """
        return cls(date)

    def load_transform(self, date):
        """ """
        try:
            from skimage import transform
        except ImportError:
            raise ImportError("Cannot import skimage. Please install scikit-image first.")
        if Time(date) < Time("2019-09-06"):
            transform_type = "Affine"
            parameters = np.asarray([[9.40402465e-01, -1.17417175e-01, -8.52095952e+02],
                                     [3.03093868e-02, -7.43615349e-01,
                                         6.94043198e+02],
                                     [0.00000000e+00,  0.00000000e+00,  1.00000000e+00]])

        elif Time(date) > Time("2019-04-17"):
            transform_type = "Affine"
            parameters = np.asarray([[8.02889540e-01, -2.30053433e-01, -6.00201420e+02],
                                     [1.94282497e-01, -5.06596412e-01,
                                      2.93073783e+02],
                                     [0.00000000e+00,  0.00000000e+00,  1.00000000e+00]])
        else:
            raise ValueError("Transformation ready for date > 2019-04-17")

        self.transformation = eval(
            "transform.%sTransform(parameters)" % transform_type)

    # ============== #
    #   Methods      #
    # ============== #
    def ccd_to_ifu(self, ccd_x, ccd_y):
        """ """
        return self.transformation(np.asarray([ccd_x, ccd_y]).T)

    def get_ifu_guess_pos(self, cube_filename=None):
        """ """
        if cube_filename is None:
            if not hasattr(self, "filename"):
                raise ValueError("no filename loaded, you must provide one.")

            cube_filename = self.filename

        return self.ccd_to_ifu(*self.get_ccd_pos(cube_filename))

    @staticmethod
    def get_ccd_pos(cube_filename):
        """ """
        return get_ccd_pos(cube_filename)


def get_astrometry_fitter(date=None, download_data=False):
    """
    See timeline here http://www.astro.caltech.edu/sedm/Hardware.html
    """
    print("***   get_astrometry_fitter IS DEPRECATED   ***")
    try:
        from ztfquery import sedm as zsedm
    except ImportError:
        raise ImportError("Cannot import ztfquery. Please install ztfquery first.")
    if date is None:
        date = "2019-09-10"

    if Time(date) > Time("2019-09-05"):
        timerange = ["2019-09-05", None]
        NON_STDTARGET = ["ZTF19abqyoxj", "ZTF19abwnpus", "ZTF19abuayqg", "ZTF19abuxhqc", "ZTF19abujuex",
                         "ZTF19abuszom", "ZTF19abyfazm", "ZTF19abwscnk", "ZTF19abqwtfu", "ZTF19abvanim",
                         "ZTF19abrlznf", "ZTF19abxvkvc", "ZTF19abuzjqa", "ZTF19abwaohs", "ZTF19abwrzqb",
                         "ZTF18abksgkt", "ZTF19abvhduf", "ZTF19abupyxe", "ZTF19abuhlxk", "ZTF19abxivpg",
                         "ZTF19abvdbyx", "ZTF19abttrte", "ZTF19ablwwox", "ZTF19aburnqh", "ZTF19abupyxe"]

        NON_STDTARGET += ["ZTF19abrdnty", "ZTF19abqykyd"]

    elif Time(date) > Time("2019-05-15"):
        print("Time Period: 2019-04-17, 2019-09-01")
        timerange = ["2019-05-15", "2019-09-01"]
        NON_STDTARGET = ["ZTF19abiahik", "ZTF19abhzelh", "ZTF19abhztqq", "ZTF19abhemuy", "ZTF19abiahik",
                         "ZTF19abhzelh", "ZTF19abidbqp", "ZTF19abgfuhh", "ZTF19abgsyrp",
                         "ZTF19abfvpkq", "ZTF19aawethv", "ZTF19abhdqlq", "ZTF19abcvrpm", "ZTF19abdrwsz",
                         "ZTF19abdzehg", "ZTF19abangub", "ZTF19abegruk", "ZTF19abdgpuv",
                         "ZTF19abbnamr", "ZTF19abdkecj", "ZTF19aatyrpj", "ZTF19aayjqpa"]

    elif Time(date) > Time("2019-04-23"):
        timerange = ["2019-04-23", "2019-05-15"]
        NON_STDTARGET = ["ZTF19aatesgp", "ZTF19aarkpif", "ZTF19aatlmbo", "ZTF19aatgxjy",
                         "ZTF19aarjfqe", "ZTF19aarstfg", "ZTF19aariwpf"]

    else:
        raise ValueError("Astrometry Fitting made for date>2019-04-17")

    # Cosmics: ZTF19abdoznh
    # ztfquery finding target files
    p = zsedm.SEDMQuery()
    if download_data:
        _ = [p.download_target_data(name, which="cube")
             for name in NON_STDTARGET]
        _ = [p.download_target_data(name, which="astrom")
             for name in NON_STDTARGET]

    target_files = p.get_local_data(NON_STDTARGET,  timerange=timerange)
    return AstrometryFitter(target_files)



#! /usr/bin/env python
# -*- coding: utf-8 -*-


""" This modules handles the CCD based images """


"""
// Full usage of ccd functionalities requires: shapely, astrobject, pynverse (all pipable) //

// ccd-x axis is refered to as 'i', y as "j".

show: plot the ccd as imshow.
show_traceindex: plot the ccd using `show()` and overplot the trace coutours.
"""


import warnings
import numpy as np

# Propobject
from propobject            import BaseObject
# Astrobject
from astrobject.photometry import Image
# PyIFU
from pyifu.spectroscopy    import Spectrum


from .utils.tools import kwargs_update

"""
The Idea of the CCD calibration is to first estimate the Spectral Matching.
Then to set this Matching to all the ScienceCCD objects.

- Tracematch holds the spaxel->pixel_area relation.

"""

__all__ = ["get_dome","get_ccd"]

##################################
#                                #
#   Object Generators            #
#                                #
##################################
def get_ccd(lampfile, ccdspec_mask=None,
            tracematch=None, background=None,
            correct_traceflexure=False, savefile_traceflexure=None,
            jscan = None,
                **kwargs):
    """ Load a SEDmachine ccd image.

    Parameters
    -----------
    lampfile: [string]
        Location of the file containing the object

    tracematch: [Tracematch (pysedm object)] -optional-
        Tracematch object containing the Spectral property on the CCD

    background: [bool] -optional-
        which kind of background do you want (i.e. background=0)
    **kwargs goes to _get_default_background_()

    Returns
    -------
    ScienceCCD (Child of CCD which is a Child of an astrobjec's Image)
    """
    lamp = ScienceCCD(lampfile, background=0)
    if tracematch is not None:
        lamp.set_tracematch(tracematch)
        # Trace Flexure Correction (if any)
        if correct_traceflexure:
            from .flexure import get_ccd_jflexure
            from .sedm    import TRACE_DISPERSION
            # Save the flexure plot
            j_offset = get_ccd_jflexure(lamp, ntraces=200, tracewidth=1,
                                        jscan=[-2.5, 2.5, 10] if jscan is None else jscan,
                                savefile=savefile_traceflexure, get_object=False)

            new_tracematch = lamp.tracematch.get_shifted_tracematch(0, j_offset)
            new_tracematch.set_buffer( TRACE_DISPERSION)
            lamp.set_tracematch(new_tracematch )
            lamp.header["JFLXCORR"] =  (True, "Is TraceMatch corrected for j flexure?")
            lamp.header["CCDJFLX"] =  (j_offset, "amplitude in pixel of the  j flexure Trace correction")
        else:
            lamp.header["JFLXCORR"] =  (False, "Is TraceMatch corrected for j flexure?")
            lamp.header["CCDJFLX"] =  (0, "amplitude in pixel of the  j flexure Trace correction")

    if background is None:
        lamp.set_background(lamp._get_default_background_(**kwargs), force_it=True)

    elif not background == 0:
        lamp.set_background(background, force_it=True)

    return lamp

def get_dome(domefile, tracematch=None,  load_sep=False, **kwargs):
    """ Load a SEDmachine domeflat image.
    (special version of get_ccd. might be moved to get_ccd...)

    Parameters
    -----------
    domefile: [string]
        Location of the file containing the object

    tracematch: [Tracematch (pysedm object)] -optional-
        Tracematch object containing the Spectral property on the CCD

    **kwargs goes to DomeCCD.__init__() if tracematch is None
             else to _get_default_background_
    Returns
    -------
     DomeCCD (Child of CCD which is a Child of an astrobjec's Image)
    """
    if tracematch is not None:
        kwargs["background"] = 0

    dome = DomeCCD(domefile, **kwargs)

    if load_sep:
        dome.datadet = dome.data/np.sqrt(np.abs(dome.data))
        dome.sep_extract(thresh=50., on="datadet")

    if tracematch is not None:
        # = Tracematch that gonna help the background
        dome.set_tracematch(tracematch)
        dome.set_background(dome._get_default_background_(**kwargs), force_it=True)

    return dome


#####################################
#                                   #
#  Raw CCD Images for SED machine   #
#                                   #
#####################################
class BaseCCD( Image ):
    """ """
    def __build__(self,bw=64, bh=64,
                      fw=3, fh=3,**kwargs):
        """ build the structure of the class

        // Doc from SEP
        bw, bh : int, -optional-
            Size of background boxes in pixels. Default is 64 [ndlr (in SEP)].

        fw, fh : int, -optional-
            Filter width and height in boxes. Default is 3 [ndlr (in SEP)].

        """

        super(BaseCCD,self).__build__(**kwargs)
        # -- How to read the image
        self._build_properties["bkgdbox"]={"bh":bh,"bw":bw,"fh":fh,"fw":fw}

    # ==================== #
    #  Internal tools      #
    # ==================== #
    def _get_sep_threshold_(self, thresh):
        """ Trick to automatically get the proper threshold for SEP extract """
        if thresh is None:
            return np.median(self.rawdata)
        return thresh

    # ==================== #
    #  Properties          #
    # ==================== #
    @property
    def data_log(self):
        return np.log10(self.data)

# ============================== #
#                                #
#  Main CCD Object               #
#  Tracing <-> CCD <-> Cube      #
#                                #
# ============================== #
class CCD( BaseCCD ):
    """ Virtual Class For CCD images that have input light """

    PROPERTIES         = ["tracematch"]
    DERIVED_PROPERTIES = ["matched_septrace_index"]

    # ------------------- #
    # Tracematch <-> CCD   #
    # ------------------- #
    def set_tracematch(self, tracematch):
        """ Attach to this instance the Tracematch """
        from .spectralmatching import TraceMatch
        if TraceMatch not in tracematch.__class__.__mro__:
            raise TypeError("The given tracematch must be a TraceMatch object")

        self._properties["tracematch"] = tracematch

    def match_trace_to_sep(self):
        """ matches the SEP ellipse with the current trace vertices.
        You must have ran sep_extract() to be able to use this method.

        (This method need Shapely.)

        You can then use the methods:
        - sepindex_to_traceindex()
        - traceindex_to_sepindex()

        Returns
        -------
        Void
        """
        # -> import and tests
        try:
            from shapely import geometry, vectorized
        except ImportError:
            raise AttributeError("Matching traces to sep requires Shapely. pip install Shapely")

        if not self.has_sepobjects():
            raise AttributeError("sep has not been ran. Do so to be able to match sep output with traces")

        # -> actual code
        x,y,a,b,theta = self.sepobjects.get(["x","y","a","b","theta"]).T
        self._derived_properties["matched_septrace_index"] =\
          {idx: np.argwhere(vectorized.contains( geometry.Polygon(self.tracematch.get_trace_vertices(idx)), x,y)).ravel()
               for idx in self.tracematch.trace_indexes}

    def sepindex_to_traceindex(self, sepindex):
        """ Give the index of an sep entry. This will give the corresponding trace index """
        if not self.has_matchedindex():
            raise AttributeError("spectral match traces has not been matched with sep indexes. Run match_tracematch_and_sep()")
        return [traceindex for traceindex, sepindexes in self.matchedindex.items() if sepindex in sepindexes]

    def traceindex_to_sepindex(self, traceindex):
        """ Give the index of an sep entry. This will give the corresponding trace index """
        if not self.has_matchedindex():
            self.match_trace_to_sep()

        return self.matchedindex[traceindex]

    def set_default_variance(self, force_it=False):
        """ define a default variance using the following formula:

        rawdata + ( median(data) - percentile(data, 16) )**2,

        it supposed to account for poisson noise + potential additional variance.
        This is a really poor's man tools...

        Returns
        -------
        Void
        """
        if self.has_var() and not force_it:
            raise AttributeError("Cannot reset the variance. Set force_it to True to allow overwritting of the variance.")
        delta_sigma = np.percentile(self.data, [16,50])

        self._properties['var'] = self.rawdata+(delta_sigma[1]-delta_sigma[0])**2

    # ----------- #
    #   GETTER    #
    # ----------- #
    def get_trace_cutout(self, traceindex, masked=True, on="data"):
        """ returns a 2D array containg the cutout around the requested trace.
        The trace could be either with or without tracematch mask (i.e. 0 outside
        the trace).

        Parameters
        ----------
        traceindex: [int]
            index of the spaxel trace.

        masked: [bool] -optional-
            do you want the data (see `on`) to be masked out (i.e. ==0) outside
            the trace

        on: [string] -optional-
            On which data source do you want the trace (e.g. data, rawdata, background, variance)

        Returns
        -------
        2d-array
        """
        xmin, xmax = self.tracematch.get_trace_xbounds(traceindex)
        ymin, ymax = self.tracematch.get_trace_ybounds(traceindex)
        if masked:
            return self.get_trace_mask(traceindex)[ymin:ymax,xmin:xmax]*eval("self.%s"%on)[ymin:ymax,xmin:xmax]
        return eval("self.%s"%on)[ymin:ymax,xmin:xmax]

    def get_finetuned_trace(self, traceindex, polydegree=2,
                            width=None, trace_position=False, **kwargs):
        """ The builds a fine tuned trace of the given traceindex.
        The tuning uses detected object from sep.
        => You must have run match_tracematch_and_sep()

        Parameters
        ----------
        traceindex: [int]
            Index of the trace you want to fine tune

        polydegree: [positive-int] -optional-
            Degree of the polynome that will be used to define the trace.
            (See 'width' for details on the width of the trace polygon)
            => If polydegree is higher than the number of sep object detected
               belonging to this trace, polydegree will the reduced to that number
            => If polydegree ends up being lower of equal to 0, None is returned

        width: [float / None] -optional-
            Width of the polygon (in pixels)
            If None, width will be estimated based on the b-values of the
            sep detected objects.

        trace_position: [bool] -optional-
            Get the expected trace central position instead of the vertices of the polygon containing it.


        **kwargs goes to spectralmatching.get_boxing_polygone()

        Returns
        -------
        One of these:
          - None if no fit possible (one of less sep object for this trace )
          - array (vertices)
          - array (x,y if trace_position=True)
        """
        if not self.has_sepobjects():
            warnings.warn("No SEP object loaded. run sep_extract() to enable finetuning. Original vertices returned")
            return self.tracematch.copy()

        _cannot_finetune = False
        try:
            x, y, b = self.sepobjects.get(["x","y","b"], self.traceindex_to_sepindex(traceindex)).T
        except:
            _cannot_finetune = True
            x,y,b = [],[],[]

        if len(x) < polydegree:
            warnings.warn("less sep-points than polynom degree. Degree reduced ")
            polydegree = len(x)-1

        if polydegree <=0:
            _cannot_finetune = True

        if  _cannot_finetune:
            warnings.warn("cannot build finetune tracing for %s. Normal vertices returned"%traceindex)
            return self.tracematch.get_trace_vertices(traceindex)
        else:
            if trace_position:
                return self.tracematch.get_finetuned_trace(traceindex, x, y, polydegree=polydegree, **kwargs)
            else:
                return self.tracematch.get_finetuned_trace_vertices(traceindex, x, y,
                                                            width= np.nanmedian(b)*2. if width is None else width,
                                                            polydegree=polydegree, **kwargs)

    def get_trace_mask(self, traceindex, finetune=False, polydegree=2,
                           subpixelisation=5, **kwargs):
        """ Build a weightmask based on the trace vertices.

        Parameters
        ----------
        traceindex: [int]
            index of the trace for which you want a mask

        finetune: [bool] -optional-
            Should the trace be remeasured based on potential detected sources?

        // The following options apply if finetune is True

        polydegree: [int] -optional-
            Degree of the polynome used to define the traces

        subpixelisation: [int] -optional-
            Our much should the pixel be subdivided to do the polygon-to-image
            interpolation? (the higher the slower)
            Set 1 for no subdivition (fastest)

        **kwargs goes to the method `get_finetuned_trace`
        """
        if not finetune:
            return self.tracematch.get_trace_mask(traceindex)

        from .spectralmatching import polygon_mask, _HAS_SKIMAGE
        if not _HAS_SKIMAGE:
            warnings.warn("get_trace needs skimage to be able to use subpixelisation")
            subpixelisation = 1

        verts = self.get_finetuned_trace(traceindex, polydegree=polydegree, **kwargs)

        mask = np.asarray(polygon_mask( verts*subpixelisation,
                                        self.shape[0]*subpixelisation, self.shape[1]*subpixelisation,
                                        get_fullcolor=False),
                              dtype="float")

        if subpixelisation==1:
            return mask/np.nanmax(mask)

        from .spectralmatching import measure
        return measure.block_reduce(mask, (subpixelisation,subpixelisation) )/float(subpixelisation**2)

    def get_finetuned_tracematch(self, indexes, polydegree=2, width=None, build_masking=False, **kwargs):
        """ """
        from .spectralmatching import TraceMatch
        tmap = TraceMatch()
        tmap.set_trace_vertices({i:self.get_finetuned_trace(i, polydegree=polydegree, width=width, **kwargs)
                                     for i in indexes}, build_masking=build_masking)
        return tmap

    # ---------------- #
    #  Tracematch      #
    # ---------------- #
    def get_spectrum(self, traceindex, on="data", finetune=False):
        """ Get the basic spectrum extracted for the CCD based on the
        TraceMatch object.

        Parameters
        ----------
        traceindex: [int, list of]
            index(es) of the spectrum(a) to return

        on: [str] -optional-
            on which 2d image shall the spectrum be extracted.
            By Default 'data', but you can set e.g. rawdata, background
            or anything accessible as 'self.%s'%on.

        finetune: [bool] -optional-
            Should the trace masking come from finetunning of spectral trace?
            (Remark: The spectral match loaded might already be finetuned ones.)

        Returns
        -------
        flux (or list of) as a function of pixels
        """
        if not self.has_tracematch():
            raise AttributeError("The TraceMatch has not been set. see set_tracematch() ")

        if hasattr(traceindex, "__iter__"):
            return [self.get_spectrum(id_) for id_ in traceindex]

        maskidx  = self.get_trace_mask(traceindex, finetune=finetune)
        return np.sum(eval("self.%s"%on)*maskidx, axis=0)

    def get_xslice(self, i, on="data"):
        """ build a `CCDSlice` based on the ith-column.

        Returns
        -------
        CCDSlice (child of Spectrum)
        """


        slice_ = CCDSlice(None)
        if "data" in on and not self.has_var():
            warnings.warn("Setting the default variance for 'get_xslice' ")
            self.set_default_variance()

        var = self.var.T[i] if 'data' in on else np.ones(np.shape("self.%s"%on))*np.nanstd("self.%s"%on)

        slice_.create(eval("self.%s.T[i]"%on), variance = var,
                    lbda = np.arange(len(self.data.T[i])), logwave=False)

        slice_.set_tracebounds(self.tracematch.get_traces_crossing_x_ybounds(i))

        return slice_

    def fit_background(self, start=2, jump=10, multiprocess=True, set_it=True, smoothing=[0,5], **kwargs):
        """ """
        from .background import get_background, fit_background
        self._background = get_background( fit_background(self, start=start, jump=jump,
                                                            multiprocess=multiprocess, **kwargs),
                                               smoothing=smoothing )
        if set_it:
            self.set_background(self._background.background, force_it=True)


    def fetch_background(self, set_it=True, build_if_needed=True, ncore=None, **kwargs):
        """ 
        ncore is used only if build_background() is called.
        """
        from .background import load_background
        from .io import filename_to_background_name
        # ---------------- #
        #  Test it exists  #
        # ---------------- #
        from glob import glob
        if len(glob(filename_to_background_name(self.filename)))==0:
            warnings.warn("No background has been found for %s"%self.filename)
            if not build_if_needed:
                raise IOError("Since build_if_needed=False, No background available.")
            from .background import build_background
            build_background(self, ncore=ncore, **kwargs)
            warnings.warn("A background has been built")

        self._background = load_background( filename_to_background_name( self.filename ))
        if set_it:
            self.set_background( self._background.background, force_it=True)

    def extract_spectrum(self, traceindex, wavesolution, lbda=None, kind="cubic",
                             get_spectrum=True, pixel_shift=0.):
        """ Build the `traceindex` spectrum based on the given wavelength solution.
        The returned object could be an pyifu's Spectrum or three arrays, lbda, flux, variance.

        The method works as follow for the given traceindex:
        1) Get the flux per pixels [using the get_spectrum() method]
           (Get the variance the same way if any)
        2) Convert the given lbda into pixels
           [using the lbda_to_pixels() method from wavesolution]
        3) Interpolate the flux per pixels into flux per lbda
           (Interpolate the variance the same way)
           [using interp1d from scipy.interpolate]


        Parameters
        ----------
        traceindex: [int]
            The index of the spectrum you want to extract

        wavesolution: [WaveSolution]
            Object containing the method to go from pixels to lbda

        lbda: [array] -optional-
            Shape of the lbda array you want the spectrum to have.

        kind: [str or int] -optional-
            Specifies the kind of interpolation as a string
            ('linear', 'nearest', 'zero', 'slinear', 'quadratic, 'cubic'
            where 'slinear', 'quadratic' and 'cubic' refer to a spline
            interpolation of first, second or third order) or as an integer
            specifying the order of the spline interpolator to use.

        get_spectrum: [bool] -optional-
            Which form the returned data should have?

        finetune: [bool] -optional-
            Should the trace masking come from finetunning of spectral trace?
            (Remark: The spectral match loaded might already be a finetuned ones.)

        Returns
        -------
        Spectrum or
        array, array, array/None (lbda, flux, variance)
        """
        f = self.get_spectrum(traceindex, finetune=False)
        v = self.get_spectrum(traceindex, finetune=False, on="var") if self.has_var() else None
        pixs = np.arange(len(f))[::-1]
        minpix, maxpix = self.tracematch.get_trace_xbounds(traceindex)
        mask = pixs[(pixs>minpix)* (pixs<maxpix)][::-1]
        if lbda is not None:
            from scipy.interpolate     import interp1d
            pxl_wanted = wavesolution.lbda_to_pixels(lbda, traceindex) + pixel_shift
            flux = interp1d(pixs[mask], f[mask], kind=kind)(pxl_wanted)
            var  = interp1d(pixs[mask], v[mask], kind=kind)(pxl_wanted) if v is not None else v
        else:
            lbda = wavesolution.pixels_to_lbda(pixs[mask], traceindex)
            flux = f[mask]
            var  = v[mask]

        if get_spectrum:
            spec = Spectrum(None)
            spec.create(flux,variance=var,lbda=lbda)
            return spec

        return lbda, flux, var

    # --------------- #
    #  Extract Cube   #
    # --------------- #
    def extract_cube(self, wavesolution, lbda,
                         hexagrid=None, traceindexes=None, show_progress=False,
                         pixel_shift=0., rotation=None):
        """ Create a cube from the ccd.

        ------------------------------------
        | Central method of the ccd object |
        ------------------------------------

        The method works as follow (see the extract_spectrum() method):
        for each trace (loop object traceindexes)
        1) Get the flux per pixels [using the get_spectrum() method]
           (Get the variance the same way if any)
        2) Convert the given lbda into pixels
           [using the lbda_to_pixels() method from wavesolution]
        3) Interpolate the flux per pixels into flux per lbda
           (Interpolate the variance the same way)
           [using interp1d from scipy.interpolate]
        4) Get the x,y position of the traceindex
           [using the ids_to_index() and index_to_xy() methods from hexagrid]

        If anything false (most likely the interpolation because of wavelength matching)
        the flux (or variance) per lbda will be set to an array of NaNs

        All the spaxels fluxes will be set to a cube (SEDMCube see .sedm)

        Parameters
        ----------
        wavesolution: [WaveSolution]
            The object containing the pixels<->wavelength relation of the night.

        lbda: [float array]
            wavelength array of the cube to be created (in Angstrom)

        hexagrid: [HexagoneProjection] -optional-
            object containing the x,y position of the traces.
            If not given, this will be created based on the instance's TraceMatch.
            (it is advised to give the night hexagrid.)

        traceindexes: [list of int] -optional-
            Which trace should be extracted to build the cube?
            If not given (None) this will used all the traceindexes for which there
            is a wavelength solution (wavesolution)

        show_progress: [bool] -optional-
            Should the progress within the loop over traceindexes should be
            shown (using astropy's ProgressBar)

        Returns
        -------
        SEDMCube (child of pyifu's Cube)
        """
        from .sedm import SEDMSPAXELS, SEDMCube, SEDM_INVERT, SEDM_ROT
        if rotation is None:
            rotation = SEDM_ROT
        # - index check
        if traceindexes is None:
            traceindexes = np.sort(list(wavesolution.wavesolutions.keys()))

        elif np.any(~np.in1d(traceindexes, list(wavesolution.wavesolutions.keys()))):
            raise ValueError("At least some given indexes in `used_indexes` do not have a wavelength solution")

        # - Hexagonal Grid
        if hexagrid is None:
            hexagrid = self.tracematch.extract_hexgrid(traceindexes)

        used_indexes = [i_ for i_ in traceindexes if i_ in hexagrid.ids_index.keys()]
        # - data
        cube      = SEDMCube(None)
        cubeflux_ = {}
        cubevar_  = {} if self.has_var() else None

        # ------------ #
        # MultiProcess #
        # ------------ #
        def _build_ith_flux_(i_):
            try:
                lbda_, flux_, variance_ = self.extract_spectrum(i_, wavesolution, lbda=lbda,
                                                                get_spectrum=False,
                                                                pixel_shift=pixel_shift)
            except:
                warnings.warn("FAILING EXTRACT_SPECTRUM for trace index %d: most likely wavesolution failed for this trace. *NaN Spectrum set*"%i_)
                flux_ = np.ones(len(lbda) )*np.NaN
                variance_ = np.ones(len(lbda) )*np.inf

            cubeflux_[i_] = flux_
            if cubevar_ is not None:
                cubevar_[i_] = variance_
            #except:
            #    warnings.warn("FAILED FOR %s. Most likely, the requested wavelength are not fully covered by the trace"%i_)
            #    cubeflux_[i_] = np.zeros(len(lbda))*np.NaN
            #    if cubevar_ is not None:
            #        cubevar_[i_] = np.zeros(len(lbda))*np.NaN

        # ------------ #
        # - MultiThreading to speed this up
        if show_progress:
            print("show_progress removed from extract_cube()")

        _ = [_build_ith_flux_(i) for i in used_indexes]

        cubeflux = np.asarray([cubeflux_[i] for i in used_indexes])
        cubevar  = np.asarray([cubevar_[i]   for i in used_indexes]) if cubevar_ is not None else None

        # - Fill the Cube
        #  SEDM DEPENDENT
        hexagrid.set_rot_degree(rotation)
        spaxels_position = np.asarray(hexagrid.index_to_xy( hexagrid.ids_to_index(used_indexes),
                                        invert_rotation=False,
                                        switch_axis=SEDM_INVERT)).T


        spaxel_map = {i:c for i,c in zip(used_indexes, spaxels_position)}

        cube.create(cubeflux.T,lbda=lbda, spaxel_mapping=spaxel_map, variance=cubevar.T)
        cube.set_spaxel_vertices(np.dot(hexagrid.grid_rotmatrix,SEDMSPAXELS.T).T)
        return cube


    # --------------- #
    #  Extract Cube   #
    # --------------- #
    def show(self, toshow="data", ax=None,
                 logscale= False, cmap = None, show_sepobjects=False,
                 vmin = None, vmax = None, savefile=None, show=True, **kwargs):
        """ Highlight the trace on top of the CCD image.
        This method requires that the spectral match has been loaded.

        Parameters
        ----------
        idx: [int]


        vmin, vmax: [float /string / None]
            Upper and lower value for the colormap. 3 Format are available
            - float: Value in data unit
            - string: percentile. Give a float (between 0 and 100) in string format.
                      This will be converted in float and passed to numpy.percentile
            - None: The default will be used (percentile 0.5 and 99.5 percent respectively).
            (NB: vmin and vmax are independent, i.e. one can be None and the other '98' for instance)

        Returns
        -------
        dict ({ax, fig, imshow's output})
        """
        from .utils.mpl import figout, mpl

        if ax is None:
            fig = mpl.figure(figsize=[8,8])
            ax  = fig.add_axes([0.13,0.13,0.8,0.8])
            ax.set_xlabel(r"$\mathrm{x\ [ccd]}$",fontsize = "large")
            ax.set_ylabel(r"$\mathrm{y\ [ccd]}$",fontsize = "large")
        elif "imshow" not in dir(ax):
            raise TypeError("The given 'ax' most likely is not a matplotlib axes. "+\
                             "No imshow available")
        else:
            fig = ax.figure

        # What To Show
        data_ = eval("self.%s"%toshow) if type(toshow) == str else toshow
        if logscale:
            data_ = np.log10(data_)

        if cmap is None:  cmap=mpl.cm.viridis
        vmin = np.percentile(data_, 0.5) if vmin is None else \
          np.percentile(data_, float(vmin)) if type(vmin) == str else\
          vmin
        vmax = np.percentile(data_, 99.5) if vmax is None else \
          np.percentile(data_, float(vmax)) if type(vmax) == str else\
          vmax

        prop = kwargs_update(dict(origin="lower", aspect='auto', vmin=vmin, vmax=vmax), **kwargs)

        # Show It
        sc = ax.imshow(data_, **prop)

        # - sepobject
        if show_sepobjects and self.has_sepobjects():
            self.sepobjects.display(ax)


        fig.figout(savefile=savefile, show=show)

        return {"ax":ax, "fig" : fig, "imshow":sc}

    def show_traceindex(self, traceindex, ax=None,
                          logscale= False, toshow="data", show_finetuned_traces=False,
                          cmap = None, facecolor = "None", edgecolor="k",
                          vmin = None, vmax = None, savefile=None, show=True, **kwargs):
        """ Highlight the trace on top of the CCD image.
        This method requires that the spectral match has been loaded.

        Parameters
        ----------
        traceindex: [int]

        """
        from .utils.mpl import figout

        pl = self.show( toshow=toshow, ax=ax, logscale=logscale, cmap = cmap,
                    vmin = vmin, vmax = vmax, savefile=None, show=False, **kwargs)
        ax, fig = pl["ax"], pl["fig"]
        pl["patch"] = self.display_traces(ax, traceindex, facecolors=facecolor, edgecolors=edgecolor)

        if show_finetuned_traces and self.has_matchedindex():
            from matplotlib import patches
            traces = traceindex if hasattr(traceindex, "__iter__") else [traceindex]
            for idx_ in traces:
                self.sepobjects.display_ellipses(pl["ax"], self.traceindex_to_sepindex(idx_))
                p_ = patches.Polygon(self.get_finetuned_trace(idx_),
                                         facecolor="None", edgecolor="C1")
                pl["ax"].add_patch(p_)


        # Output
        fig.figout(savefile=savefile, show=show)
        return pl

    def show_xslice(self, xpixel, toshow="data", savefile=None, ax=None, show=True,
                        ls="-", color=None, lw=1,
                        show_tracebounds=True,
                        bandalpha=0.5, bandfacecolor="0.7", bandedgecolor="None",bandedgewidth=0,
                        **kwargs):
        """ """
        from .utils.mpl import figout, mpl

        if ax is None:
            fig = mpl.figure(figsize=[9,5])
            ax  = fig.add_subplot(111)
        else:
            fig = ax.figure

        # This plot
        ax.plot( eval("self.%s.T[xpixel]"%toshow), ls=ls, color=color, lw=lw,**kwargs)
        if show_tracebounds:
            [ax.axvspan(*y_, alpha=bandalpha, facecolor=bandfacecolor,
                            edgecolor=bandedgecolor,linewidth=bandedgewidth)
                 for y_ in self.tracematch.get_traces_crossing_x_ybounds(xpixel)]

        fig.figout(savefile=savefile, show=show)


    def show_as_slicer(self, traceindexes, vmin=0 , vmax="90",
                           masked=True, toshow="rawdata"):
        """ """
        import matplotlib.pyplot as mpl
        data = eval("self.%s"%toshow)
        if vmin is None:
            vmin = 0
        if type(vmin) == str:
            vmin = np.percentile(data, vmin)
        if vmax is None:
            vmax = "95"
        if type(vmax) == str:
            vmax = np.percentile(data, vmax)

        #  parameters
        ntraces = len(traceindexes)
        height = 0.85/ntraces

        #  Build the figure
        fig = mpl.figure(figsize=[8,4])

        #  draw the traces
        for i,index_ in enumerate(traceindexes):
            ax  = fig.add_axes([0.1,0.1+height*i, 0.8,height])
            ax.imshow(self.get_trace_cutout(index_, masked=masked, on=toshow), origin="lower", aspect="auto", vmin=vmin, vmax=vmax)
            ax.set_yticks([])
            if i>0:
                ax.set_xticks([])
            else:
                ax.set_xlabel("pixels since trace origin")

        return {"fig":fig}

    def display_traces(self, ax, traceindex, facecolors="None", edgecolors="k",
                             update_limits=True):
        """ """
        pl = self.tracematch.display_traces(ax, traceindex,
                                            facecolors =facecolors, edgecolors=edgecolors)

        # Fancy It
        if update_limits:
            [xmin, xmax], [ymin, ymax] = np.percentile(self.tracematch.trace_vertices[traceindex] if not hasattr(traceindex,"__iter__")  else\
                                                       np.concatenate([self.tracematch.trace_vertices[idx] for idx in traceindex], axis=0),
                                                       [0,100], axis=0).T
            ax.set_xlim(xmin-20, xmax+20)
            ax.set_ylim(ymin-5, ymax+5)

        return pl

    # ================== #
    #  Internal Tools    #
    # ================== #
    def _get_default_background_(self, add_mask=None,
                                     cut_bright_pixels=None,
                                     exclude_edges=False,
                                 scaleup_sepmask=2, apply_sepmask=True,
                                 **kwargs):
        """ This Background has been optimized for SEDm Calibration Lamps """

        if add_mask is None and self.has_tracematch():
            add_mask = np.asarray(~self.tracematch.get_notrace_mask(), dtype="bool")

        if add_mask is not None and cut_bright_pixels is not None:
            data_ = self.rawdata.copy()
            data_[add_mask] = np.NaN
            add_mask = add_mask + (data_>np.percentile(data_[data_==data_],50))

        if exclude_edges:
            falses = np.zeros(self.shape)
            # - x cuts
            xremove, yremove =100, 20
            falses[:,(np.arange(self.shape[1])<xremove) + (np.arange(self.shape[1])>(self.shape[1]-xremove))  ] = 1.
            falses[(np.arange(self.shape[0])<yremove)   + (np.arange(self.shape[0])>(self.shape[0]-yremove)),:] = 1.
            add_mask = add_mask + np.asarray(falses, dtype="bool")

        return self.get_sep_background(doublepass=False, update_background=False,
                                       add_mask=add_mask,
                                       apply_sepmask=apply_sepmask, scaleup_sepmask=scaleup_sepmask,
                                       **kwargs)

    # ================== #
    #   Properties       #
    # ================== #

    # - TraceMatching association
    @property
    def tracematch(self):
        """ """
        return self._properties["tracematch"]

    def has_tracematch(self):
        return self.tracematch is not None

    @property
    def matchedindex(self):
        """ Object containing the relation between the sepindex and the trace index.
        see the methods sepindex_to_traceindex() and traceindex_to_sepindex() """
        if self._derived_properties["matched_septrace_index"] is None:
            self._derived_properties["matched_septrace_index"] = {}
        return self._derived_properties["matched_septrace_index"]

    def has_matchedindex(self):
        """ Is the sep<-> trace index matching done? """
        return self.matchedindex is not None and len(self.matchedindex.keys())>0

    # - Generic properties
    @property
    def objname(self):
        if "Calib" in self.header.get("NAME","no-name"):
            return self.header["NAME"].split()[1]
        return self.header.get("NAME","no-name")

# ============================== #
#                                #
#  Childs Of CCD                 #
#                                #
# ============================== #
class ScienceCCD( CCD ):
    """ Should be used to improve the trace matching. """


class DomeCCD( ScienceCCD ):
    """ Object Build to handle the CCD images of the Dome exposures"""

    # ================== #
    #   Main Tools       #
    # ================== #
    def get_tracematch(self, bound_pixels=None, width="optimal"):
        """ """
        xlim, ylim = self.get_trace_position(bound_pixels=bound_pixels)

        if width is None:
            b = self.sepobjects.get("b")
            width = np.median(b)*2 if width=="median" else \
              b*2 if width!="optimal" else np.clip(b,np.median(b)-2*np.std(b),np.median(b)+2*np.std(b))*2


        return  [np.concatenate(v_) for v_ in zip(np.asarray([xlim,ylim+width]).T,np.asarray([xlim[::-1],ylim[::-1]-width]).T)]


    def get_trace_position(self, bound_pixels=None):
        """ """
        x, y, a, b, theta = self.sepobjects.get(["x","y","a","b","theta"]).T
        # This enables to remove cosmics
        flagout =  a/b<10
        x, y, theta = x[~flagout], y[~flagout], theta[~flagout]

        if bound_pixels is not None:
            print("BOUNDS")
            xlim = np.asarray([x-bound_pixels[0], x+bound_pixels[1]])
            ylim = np.sin(theta)*np.asarray([[-bound_pixels[0], bound_pixels[1]]]).T + y
            return xlim, ylim

        from .sedm import domexy_to_tracesize
        return domexy_to_tracesize(x, y, theta)



    def get_specrectangles(self, length="optimal",height="optimal",theta="optimal",
                               scaleup_red=2.5, scaleup_blue = 7.5, scaleup_heigth=1.5,
                               fixed_pixel_length=None,
                               use_peak_xy=True):
        """ Vertices of the ~rectangles associated to each spectra based on the
        detected SEP ellipses assuming a fixed height given by the median of the
        b-ellipse parameter

        To convert e.g. the i-th vertice into a Shapely polygon:
        ```
        from shapely import geometry
        rect_vert = self.get_specrectangles()
        rect = geometry.polygon.Polygon(rect_vert.T[i].T)
        ```

        Parameters
        ----------

        fixed_pixel_length: [float/None] -optional-
            how many pixel bluer (higher-x) should the traces be in comparison to the
            reddest point (defined by scaleup_red).
            If This is not None, scaleup_blue is ignored
        Returns
        -------
        ndarray (4x2xN) where N is the number of detected ellipses.
        """
        if not self.has_sepobjects():
            self.sep_extract()

        x, y, xpeak, ypeak, a, b, theta = self.sepobjects.get(["x","y","xpeak", "ypeak", "a", "b", "theta"]).T
        x_,y_ = (xpeak, ypeak) if use_peak_xy else (x,y)


        length = np.median(a) if length=="median" else \
          a if length!="optimal" else np.clip(a,np.median(a)-2*np.std(a),np.median(a)+2*np.std(a))

        height = np.median(b)*scaleup_heigth if height=="median" else \
          b*scaleup_heigth if height!="optimal" else np.clip(b,np.median(b)-2*np.std(b),np.median(b)+2*np.std(b))*scaleup_heigth

        angle  = np.median(theta) if theta=="median" else \
          theta if theta!="optimal" else np.clip(theta,np.median(theta)-1.5*np.std(theta),
                                                     np.median(theta)+1.5*np.std(theta))


        left_lim = np.asarray([  x_-length*np.cos(-angle)*scaleup_red,
                                 y_+length*np.sin(-angle)*scaleup_red])
        if fixed_pixel_length is None:
            right_lim = np.asarray([ x_+length*np.cos(-angle)*scaleup_blue,
                                    y_-length*np.sin(-angle)*scaleup_blue])
        else:
            right_lim = np.asarray([ left_lim[0] + np.cos(-angle)*fixed_pixel_length,
                                     left_lim[1] - np.sin(-angle)*fixed_pixel_length])

        return np.asarray([[left_lim[0],left_lim[0],right_lim[0],right_lim[0]],
                               [left_lim[1]-height,left_lim[1]+height,right_lim[1]+height,right_lim[1]-height]]).T


    # ================== #
    #   Internal         #
    # ================== #
    def _get_default_background_(self, mask_prop={},
                                     from_spectmatch=True,
                                apply_sepmask=False, **kwargs):
        """ This Background has been Optimized for SEDm Dome """

        if from_spectmatch and self.has_tracematch():
            return super(DomeCCD, self)._get_default_background_(**kwargs)

        self.set_background(np.percentile(self.rawdata,0.001), force_it=True)
        self.sep_extract(thresh=5, err=np.percentile(self.rawdata,1))

        return self.get_sep_background(doublepass=False, update_background=False,
                                       apply_sepmask=True, **kwargs)


    def _get_sep_extract_threshold_(self):
        """this will be used as a default threshold for sep_extract"""
        warnings.warn("_get_sep_extract_threshold_ called")
        if not hasattr(self,"_sepbackground"):
                _ = self.get_sep_background(update_background=False)
        return self._sepbackground.globalrms*2


# ============================== #
#                                #
#  Main CCD Object               #
#  Tracing <-> CCD <-> Cube      #
#                                #
# ============================== #
class CCDSlice( Spectrum ):
    """ Flux per Pixel Spectrum based on CCD Slicing"""
    PROPERTIES = ["tracebounds"]
    DERIVED_PROPERTIES = ["tracemasks", "contmodel", "flagin"]

    def fit_continuum(self, degree=5, legendre=True, clipping=[2,2],
                          ngauss=0):
        """ """
        from astropy.stats import mad_std
        from modefit import get_polyfit, get_normpolyfit
        y = self.data[self.tracemaskout]

        nmad = mad_std(y[y==y])
        median = np.nanmedian(y)
        self._derived_properties["flagin"] = ((median - clipping[0] * nmad < y) &  (y< median + clipping[1] * nmad)) & (y==y)

        guesses = dict(a0_guess=np.nanmedian(self.data[self.tracemaskout][self.flagin]))
        if ngauss > 0:
            self._derived_properties["contmodel"] = \
              get_normpolyfit(self.lbda[self.tracemaskout][self.flagin], self.data[self.tracemaskout][self.flagin],
                            np.sqrt(self.variance[self.tracemaskout][self.flagin]), degree, ngauss=ngauss, legendre=legendre)
            for ni in range(ngauss):
                guesses['mu%d_guess'%ni]       = np.argmax(self.data)
                guesses['mu%d_boundaries'%ni]  = [0, len(self.data)]
                guesses['sig%d_guess'%ni]      = 150
                guesses['sig%d_boundaries'%ni] = [100,200]
                guesses['ampl%d_guess'%ni]     = np.nanmax(self.data)
                guesses['ampl%d_boundaries'%ni]= [0,None]
        else:
            self._derived_properties["contmodel"] = \
              get_polyfit(self.lbda[self.tracemaskout][self.flagin], self.data[self.tracemaskout][self.flagin],
                           np.sqrt(np.abs(self.variance[self.tracemaskout][self.flagin])), degree, legendre=legendre)

        self.contmodel.use_minuit = True
        self.contmodel.fit(**guesses)


    def show_continuum(self, ax=None, savefile=None, show=True, show_model=True):
        """ """
        from .utils.mpl  import figout, mpl
        from pyifu.tools import specplot

        if ax is None:
            fig = mpl.figure(figsize=[8,4])
            ax  = fig.add_subplot(111)
            ax.set_xlabel(r"$\mathrm{x\ [ccd]}$",fontsize = "large")
            ax.set_ylabel(r"$\mathrm{y\ [ccd]}$",fontsize = "large")
        else:
            fig = ax.figure

        ax.specplot(self.lbda, self.data, var=self.variance, color="0.7")
        dataout = self.data.copy()
        dataout[~self.tracemaskout] = np.NaN
        ax.specplot(self.lbda, dataout, var=None, color="C1")

        if show_model and self.contmodel is not None:
            ax.plot(self.contmodel.xdata, self.contmodel.data,
                        color="C0", zorder=6, ls="None", marker="x",  ms=4)
            ax.plot(self.lbda[self.tracemaskout][self.flagin], self.contmodel.model.get_model(),
                        color="C0", zorder=7)

        fig.figout(savefile=savefile, show=show)

    def get_fit_contrains(self, fixed_pos=True,
                              sigma_guess=1.5,
                              pos_bounds=[1,1], sigma_bounds=[1.2,1.8], **kwargs):
        """ """
        expected_lines = self.get_expected_lines()

        fitprop = {}
        for i in range(self.ntraces):
            fitprop["mu%d_guess"%i]         = expected_lines[i]
            fitprop["mu%d_fixed"%i]         = fixed_pos
            fitprop["mu%d_boundaries"%i]    = [expected_lines[i]-pos_bounds[0],expected_lines[i]+pos_bounds[0]]
            fitprop["sig%d_guess"%i]        = sigma_guess
            fitprop["sig%d_fixed"%i]        = self.data[int(expected_lines[i])]*sigma_guess*2.5 < 200
            fitprop["sig%d_boundaries"%i]   = sigma_bounds
            fitprop["ampl%d_guess"%i]       = np.max([0,self.data[int(expected_lines[i])]*sigma_guess*2.5])
            fitprop["ampl%d_fixed"%i]       = False
            fitprop["ampl%d_boundaries"%i]  = [0,fitprop["ampl%d_guess"%i]*2 if fitprop["ampl%d_guess"%i] > 0 else 1e-4]

        return fitprop

    def get_expected_lines(self):
        """ """
        return np.mean(self.tracebounds, axis=1)

    # =================== #
    #   Properties        #
    # =================== #
    @property
    def tracebounds(self):
        """ boundaries of each traces overlapping with the selected ccd-slice """
        return self._properties["tracebounds"]

    def set_tracebounds(self, tracebounds):
        """ attach to this spectrum boundaries of each traces overlapping with the selected ccd-slice """
        self._properties["tracebounds"] = np.asarray(tracebounds, dtype="float")
        self._derived_properties["tracemasks"] = [(b[0]<self.lbda) * (self.lbda<b[1]) for b in tracebounds]

    @property
    def ntraces(self):
        """ number of traces the Slice is overlapping """
        return len(self.tracebounds)

    @property
    def tracemasks(self):
        """ boolean arrays indicating pixels overlaping spaxel traces """
        return self._derived_properties["tracemasks"]

    @property
    def tracemaskout(self):
        """ Boolean array been True for pixels not within any trace """
        return ~np.asarray(np.sum(self.tracemasks, axis=0), dtype="bool")

    @property
    def flagin(self):
        """ boolean array indicating the data point used for the fit. """
        return self._derived_properties["flagin"]

    # Model
    @property
    def contmodel(self):
        """ Continuum model measured when running fit_continuum """
        return self._derived_properties["contmodel"]



# ============================ #
#                              #
#    Background                #
#                              #
# ============================ #



#! /usr/bin/env python
#

""" Module to pull pysedm data """


try:
    import ztfquery
    from ztfquery import sedm
except ImportError:
    raise ImportError("Please install ztfquery (v>1.6.0). pip install ztfquery")

# ================= #
#    Main object    #
# ================= #
SEDMQ = sedm.SEDMQuery()


# ======================== #
#                          #
#         MAIN             #
#                          #
# ======================== #
if  __name__ == "__main__":
    import argparse
    import numpy as np
    
    parser = argparse.ArgumentParser(
             description=""" pull SEDM data from pharos to your computer""",
             formatter_class=argparse.RawTextHelpFormatter)

    # - Infile - #
    parser.add_argument('infile', type=str, default=None,
                        help='Target Name')

    # ---------- #
    #  Options   #    
    # ---------- #
    parser.add_argument('--timerange',  type=str, default="2018-08-01,now",
                        help='You only one sedm data taken during that time range.'+"\n"+\
                             'format: coma separated values "a,b" \n Could be: '+"\n"+\
                             '- 1 time means starting point (here after the 12th of Sept 2018):  --timerange 2018-09-12  '+"\n"+\
                             '- 2 times means boundaries  (here between 12th and 15th of Sept 2018):  --timerange 2018-09-12,2018-09-15'+"\n"+\
                             '- for only upper limit  (here before 15th of Sept 2018):  --timerange None,2018-09-15'
                            )

    parser.add_argument('--update',  action="store_true", default=False,
                        help='Update the sedm local database of what have been observed (what-files)'
                            )

    parser.add_argument('--data',  type=str, default="cube,spec",
                        help='Which kind of data do you want?'+"\n"+\
                             'format: coma separated values "cube,spec,crr"'+"\n"+
                             "NB: cube=e3d files ; spec = spectrum files ; ccd = ccd files "
                            )

    parser.add_argument('--fluxcal',  action="store_true", default=False,
                        help='Do you want to download the fluxcal files of the nights with your target?'
                            )
        
    parser.add_argument('--output',  type=str, default="default",
                        help='Where should the files be downloaded? (here provide the full path of the directory)'+"\n"+\
                             "by default, it will be in your $ZTFDATA/SEDM/redux/DATES directories following pharos' structure"
                            )

    parser.add_argument('--extension',  type=str, default="*",
                        help='Which extension should be file have?'+"\n"+\
                             'format: coma separated values "pdf,fits,png"'
                            )
    
    parser.add_argument('--overwrite',  action="store_true", default=False,
                        help='If the file you want to download already exists, shall this overwrite it?'
                            )
    
    parser.add_argument('--nodl',  action="store_true", default=False,
                        help='Use this option to not download (only see what should be downloaded where)'
                            )
    
    parser.add_argument('--nprocess',  type=int, default=None,
                        help='How many parallel downloading ?'
                            )
    
    parser.add_argument('--quiet', action="store_true", default=False,
                        help='This prints information. Should this be quite?'
                            )
    
    args   = parser.parse_args()
    
    # ---------- #
    #  Target    #    
    # ---------- #
    if args.update:
        if not args.quiet:
            print("Updating sedmdata")
        SEDMQ.update_sedmdata()
        

    
    target = args.infile
    if not args.quiet:
        print(" pysedm pull ".center(40,'='))
        print("Downloading data for target: %s"%target)
        print("credit: based on ztfquery")
        
    # ---------- #
    #  Script    #    
    # ---------- #
    
    # Downloading options
    dl_options = dict(nodl=args.nodl,download_dir=args.output,
                     nprocess=args.nprocess, verbose=args.quiet,
                     overwrite=args.overwrite)

    # Actual Download function for pysedm files associated to the target
    url, localdl = SEDMQ.download_target_data(target, which=args.data.split(","),
                                   timerange=args.timerange.split(",") if args.timerange is not None else [None,None],
                                   extension=args.extension if not "," in args.extension else args.extension.split(","),
                                   **dl_options
                                )
    # Downloading associated fluxcalibration files if requested
    if args.fluxcal:
        url_fluxcal,localdl_fluxcal = [], []
        target_nights = np.unique(SEDMQ.get_target_data(target)["night"])
        if not args.quiet:
            print("info: downloading fluxcalibration files for night(s): "+", ".join(target_nights))
            
        for target_night in target_nights:
            url_fluxcal_,localdl_fluxcal_ = SEDMQ.download_night_fluxcal(target_night, **dl_options)
            url_fluxcal    +=url_fluxcal_
            localdl_fluxcal+=localdl_fluxcal_
                                           
        url    += url_fluxcal
        localdl+= localdl_fluxcal

    # ---------- #
    #  Outputs   #    
    # ---------- #
    if args.nodl:
        if not args.quiet:
            print(" no download option (--nodl) ".center(40,'='))
            print(" Data to be downloaded ".center(40,'-'))
            print("\n".join(url))
            print(" Download location ".center(40,'-'))
            print("\n".join(localdl))
            

    


    



#! /usr/bin/env python
#

DESCRIPTION = "pysedm: the SEDmachine pipleine"
LONG_DESCRIPTION = """ Pipeline for generic Transient Typer IFU - option on SEDM built in. """

DISTNAME = 'pysedm'
AUTHOR = 'Mickael Rigault'
MAINTAINER = 'Mickael Rigault' 
MAINTAINER_EMAIL = 'mickael.rigault@clermont.in2p3.fr'
URL = 'https://github.com/MickaelRigault/pysedm/'
LICENSE = 'BSD (3-clause)'
DOWNLOAD_URL = 'https://github.com/MickaelRigault/pysedm/tarball/0.30'
VERSION = '0.31.0'

try:
    from setuptools import setup, find_packages
    _has_setuptools = True
except ImportError:
    from distutils.core import setup
    _has_setuptools = False

if __name__ == "__main__":

    if _has_setuptools:
        packages = find_packages()
        print(packages)
    else:
        # This should be updated if new submodules are added
        packages = ['pysedm',
                    "pysedm.script",
                    "pysedm.utils"]

    setup(name=DISTNAME,
          author=AUTHOR,
          author_email=MAINTAINER_EMAIL,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          long_description=LONG_DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=VERSION,
          download_url=DOWNLOAD_URL,
          install_requires=[
                "astropy>=4.0.5",
                "dask>=2023.2.1",
                "iminuit>=2.0.0",
                "matplotlib",
                "modefit>=0.4.0",
                "numpy>=1.21.6",
                "pandas>=1.3.0",
                "psfcube>=0.9.0",
                "pyifu",
                "scipy>=0.16.0",
                "shapely>=1.8.1",
          ],
          extra_requires=[
                "astrobject",
                "photoifu",
                "pycalspec",
                "scikit-image",
                "sep",
                "ztfquery>=1.6.0",
          ],
          scripts=["bin/ccd_to_cube.py" ,
                   "bin/extract_star.py",
                   "bin/extractstar.py",
                   "bin/extractstar_ab.py",
                   "bin/cube_quality.py",
                   "bin/pysedm_report.py",
                   "bin/pysedm_pull.py",
                   "bin/pysedm_io.py",                   
                   "bin/derive_wavesolution.py",
                   "bin/quality_check.py",
                   "bin/build_guider.py",
                   "bin/rm_cubecr.py"],
          packages=packages,
          include_package_data=True,
          package_data={'pysedm': ['data/*.*']},
          classifiers=[
              'Intended Audience :: Science/Research',
              'Programming Language :: Python :: 2.7',
              'Programming Language :: Python :: 3.5',              
              'License :: OSI Approved :: BSD License',
              'Topic :: Scientific/Engineering :: Astronomy',
              'Operating System :: POSIX',
              'Operating System :: Unix',
              'Operating System :: MacOS'],
          )



#! /usr/bin/env python
# -*- coding: utf-8 -*-


import numpy as np
import warnings
from propobject import BaseObject


"""
j-offset flexure tool. "J-offsets" (perpendicular to trace dispersion) lead to lower signal to noise (more background, less signal).

j-offset is measured by maximizing the total flux within the randomly selected subsample of traces while shifting the traces up and down.
"""

########################
#                      #
#  Optimize            #
#                      #
########################
def get_ccd_jflexure(ccd, ntraces=100, tracewidth=1,
                         jscan=[-2.5,2.5,10], savefile=None, get_object=False):
    """ give a ccd object (with tracematch loaded) ; this estimate the ccd-j trace flexure.
    [takes about ~5s]

    Parameters:
    -----------
    ccd: [pysedm.CCD]
       ccd object 

    ntraces: [int] -optional-
        Number of randomly selected traces used for the estimation.

    tracewidth: [float] -optional-
        Width (buffer) around trace lines. Pixels contained within line position +/- width are considered as within the trace. 

    jscan: [float, float, int] -optional-
        start, end and step to be scanned in j (jscan-> np.linspace(*jscan))

    savefile: [string/None] -optional-
        If you want to save the figure, provide here it's name.
        If could be a list of filename (if you want several extention for instance)

    get_object: [bool] -optional-
        By default the function returns the j-shift (float).
        However, You can get the full TraceFlexureFit object by setting 'get_object' to True

    Returns
    -------
    flaot (ccd-j shift to apply) [or TraceFlexureFit, see get_object option]
    """
    from .sedm import INDEX_CCD_CONTOURS
    smap = ccd.tracematch.get_sub_tracematch( np.random.choice(ccd.tracematch.get_traces_within_polygon( INDEX_CCD_CONTOURS), ntraces) )
    smap.set_buffer(tracewidth)
    jflex = TraceFlexureFit(ccd, smap)
    jflex.build_pseudomag_scan(*jscan)
    if savefile is not None:
        for savefile_ in np.atleast_1d(savefile):
            jflex.show(savefile=savefile_)
        
    return jflex.estimate_jshift() if not get_object else jflex


class TraceFlexureFit( BaseObject ):
    """ """
    PROPERTIES = ["ccd", "tracematch"]
    DERIVED_PROPERTIES = ["pseudomag_scan"]
    def __init__(self, ccd, tracematch):
        """ """
        self._properties["ccd"] = ccd
        self._properties["tracematch"] = tracematch

    # ===================== #
    #   Method              #
    # ===================== #
    def get_total_pseudomag(self, jshift, subpixelization=2):
        """ """
        smap_current = self.tracematch.get_shifted_tracematch(0,jshift)
        weightmap = smap_current.get_traceweight_mask(subpixelization)
        return -np.log10( np.nansum(self.ccd.data*weightmap) )

        

    def fmin_jshift(self, guess):
        return fmin(self.get_total_pseudomag, guess)

    def build_pseudomag_scan(self, start=-3, end=3, step=10):
        """ """
        from scipy.interpolate import interp1d
        self.pseudomag_scan["js"] = np.linspace(start, end, step)
        self.pseudomag_scan["pseudomag"] = [self.get_total_pseudomag(jshift_) for jshift_ in self.pseudomag_scan["js"]]
        self.pseudomag_scan["interpolation"] = interp1d(self.pseudomag_scan["js"], self.pseudomag_scan["pseudomag"],
                                                            kind="cubic")
    def estimate_jshift(self, nbins=100):
        """ """
        xx = np.linspace(self.pseudomag_scan["js"].min(), self.pseudomag_scan["js"].max(), 100)
        return xx[np.argmin(self.pseudomag_scan["interpolation"](xx))]

    # ------------- #
    #  PLOTTER      #
    # ------------- #
    def show(self, savefile=None, ax=None):
        """ """
        import matplotlib.pyplot as mpl
        if ax is None:
            fig = mpl.figure(figsize=[5,3])
            ax  = fig.add_axes([0.15,0.2,0.75,0.7])
        else:
            fig = ax.figure

        ax.plot( self.pseudomag_scan["js"], self.pseudomag_scan["pseudomag"], marker="o", label="data")
        xx = np.linspace(self.pseudomag_scan["js"].min(), self.pseudomag_scan["js"].max(),100)
        ax.plot(xx, self.pseudomag_scan["interpolation"](xx), color="C1", label="interpolation")
        
        # Result
        jshift = self.estimate_jshift()
        ax.axvline( jshift, color="0.5" )
        ax.text(jshift-0.1, np.max(self.pseudomag_scan["pseudomag"]), "j-flexure: %+.1f pixels"%jshift,
                    rotation=90, color="0.5", va="top", ha="right")

        
        # - labels
        ax.set_ylabel("Total CCD pseudo magnitude", fontsize="medium")
        ax.set_xlabel("ccd-j shift", fontsize="medium")
        ax.legend(loc="lower right", fontsize="small")
        if savefile is not None:
            fig.savefig(savefile, dpi=250)
        else:
            return {"ax":ax, "fig":fig}
        
    
    # ===================== #
    #   Properties          #
    # ===================== #
    @property
    def ccd(self):
        """ """
        return self._properties["ccd"]

    @property
    def tracematch(self):
        """ """
        return self._properties["tracematch"]

    @property
    def pseudomag_scan(self):
        """ """
        if self._derived_properties['pseudomag_scan'] is None:
            self._derived_properties['pseudomag_scan'] = {}
        return self._derived_properties['pseudomag_scan']
########################
#                      #
#  Class               #
#                      #
########################

class TraceFlexure( BaseObject ):
    """ """
    PROPERTIES = ["ccd", "mapper"]
    DERIVED_PROPERTIES = ["flagsep_used", "js", "js_obs"]

    def __init__(self, ccd, mapper=None):
        """ """
        print("DEPRECATED USED TraceFlexureFit")
        self.set_ccd(ccd)
        if mapper is not None:
            self.set_mapper(mapper)
        
    # ================== #
    #   Properties       #
    # ================== #
    def derive_j_offset(self, var_threshold=5, elliptical_limit=0.5, rerun_sepextract=False, verbose=False):
        """ """
        if not self.ccd.has_var():
            if verbose: print("INFO: setting the default variance to the ccd")
            self.ccd.set_default_variance()
        if not self.ccd.has_sepobjects() or rerun_sepextract:
            try:
                if verbose: print("INFO: Running  sep_extract")
                self.ccd.sep_extract()
            except:
                tresh = np.nanmean( np.sqrt(self.ccd.var) )*var_threshold
                if verbose: print("INFO: FAILED running sep_extract. Trying now using tresh=%.2f"%tresh)
                self.ccd.sep_extract(tresh)
                
        # - The SEP data
        a,b,x,y = self.ccd.sepobjects.get(["a","b",'x','y']).T
        self._derived_properties['flag_sepused'] = ((1- b/a)<elliptical_limit)

        # - Getting the expected J
        i_obs, j_obs = x, y
        if verbose: print("INFO: mesuring the expected j-coordinates")
        js= []
        for i_,j_ in zip(x, y):
            try:
                js.append( self.mapper.get_expected_j(i_,j_) )
            except:
                js.append(np.NaN)

        
        self._derived_properties['js']     = np.asarray(js)
        self._derived_properties['js_obs'] = np.asarray(j_obs)
        self._derived_properties['flag_sepused'] = ((1- b/a)<elliptical_limit) * ( self.js==self.js )
        
        if verbose: print("INFO: j-offset found: %+.2f pixels"%self.j_offset)
        return self.j_offset
            
    # ----------- #
    #  SETTER     #
    # ----------- #
    def set_ccd(self, ccd_):
        """  """
        self._properties["ccd"] = ccd_
        
    def set_mapper(self, mapper_):
        """ attach to this intance a mapper object containing the 
        ccd to cube coordinates conversions """
        self._properties["mapper"] = mapper_

    # ----------- #
    #  PLOTTER    #
    # ----------- #
    def show_j_flexure_ccd(self, show_sepobjects=False, zoomin=[600, 1800, 1300, 1350],
                             show=True, savefile=None):
        """ """
        import matplotlib.pyplot as mpl
        if self._derived_properties['js'] is None:
            _NO_JS = True
            warnings.warn("j offset not measured yet. see derive_j_offset()")
        else:
            _NO_JS = False

        
        fig    = mpl.figure(figsize=[10,3.5])
        axccd  = fig.add_axes([0.36,0.15,0.61,0.8])
        axhist = fig.add_axes([0.03,0.15,0.25,0.8])
        axccd.set_xlabel("i [ccd coordinates in pixels]", fontsize="large")
        axccd.set_ylabel("j [ccd coordinates in pixels]", fontsize="large")
        
        pl = self.ccd.show(ax=axccd, show=False)
        self.mapper.tracematch.display_traces(axccd, self.mapper.traceindexes, edgecolors="0.7", linestyle="--")
        self.ccd.tracematch.display_traces(axccd, self.mapper.traceindexes, edgecolors="w")
        
        if show_sepobjects:
            ells = self.ccd.sepobjects.get_detected_ellipses(scaleup=2.5, mask=self._flag_sepused,contours=False)
            for ell in ells:
                axccd.add_artist(ell)
                ell.set_clip_box(axccd.bbox)
                ell.set_facecolor("None")
                ell.set_edgecolor("r")
        if not _NO_JS:
            axhist.hist(self.delta_js[self.delta_js==self.delta_js])
            axhist.axvline(self.j_offset, ls="--",color="k")
            axhist.text(self.j_offset, 0.5, "j trace offset %+.2f pixels"%self.j_offset,
                    va="bottom",ha="right", rotation=90)
            axhist.set_xlabel(r"$\Delta$ j for %d ellipses"%(len(self._flag_sepused[self._flag_sepused])),
                         fontsize="large")
        else:
            axhist.text(0.5, 0.5, "j offset not measured yet. \n see derive_j_offset()",
                    va="center",ha="center", rotation=45, transform=axhist.transAxes)
            
        axccd.figure.canvas.draw()
        axccd.set_xlim(*zoomin[:2])
        axccd.set_ylim(*zoomin[2:])
        if show:
            fig.show()
        if savefile:
            fig.savefig(savefile)
    # ================== #
    #   Properties       #
    # ================== #
    @property
    def ccd(self):
        """ the ccd object """
        return self._properties["ccd"]
    @property
    def mapper(self):
        """ object containing the ccd to 3d cube conversion tools """
        return self._properties["mapper"]

    # - Derived properties
    @property
    def js_obs(self):
        """ the observed j positions """
        if self._derived_properties['js_obs'] is None:
            raise AttributeError("This has not been set. see derive_j_offset()")
        return self._derived_properties['js_obs']

    @property
    def js(self):
        """ the expected j positions """
        if self._derived_properties['js'] is None:
            raise AttributeError("This has not been set. see derive_j_offset()")
        return self._derived_properties['js']
    
    @property
    def delta_js(self):
        """ Offset between the expected j position and the observed onces """
        return self.js_obs - self.js
    
    @property
    def j_offset(self):
        """ """
        return np.nanmedian(self.delta_js)
    
    @property
    def _flag_sepused(self):
        """ """
        return self._derived_properties['flag_sepused']



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import pysedm
import datetime
from astropy.io import fits

    
def build_image_report(specfile):
    """ """
    from pysedm.utils import pil
    now = datetime.datetime.now()
    timestring = "made the %d-%02d-%02d at %02d:%02d:%02d"%(now.year,now.month,now.day, now.hour, now.minute, now.second)
        
    footer = pil.get_buffer([20,1], "pysedm version %s"%pysedm.__version__+ "  |  " + timestring,
                                fontsize=15, textprop=dict(color="0.5"), barcolor="k")

    header  = fits.getheader(specfile)
    date    = pysedm.io.header_to_date(header)
    spec_id = pysedm.io.filename_to_id(specfile)
    filesourcename = specfile.split("spec_")[-1].split(".fits")[0]
    if "+" in filesourcename:
        filesourcename = filesourcename.split("+")[0]
        
    object_name = header['OBJECT'].split()[0]  # remove the [A] in 'TARGET [A]'
    
    if "STD" in filesourcename:
        STD = True
        filesourcename = filesourcename.split("STD")[0]
    else:
        STD = False
        
    # Extraction Mode    
    extraction_mode = header["EXTRTYPE"] if "EXTRTYPE" in header else "auto"
    
    # Missing plot format
    prop_missing = dict(fontsize=30, textprop=dict(color="C1"))

    # Flexure Images
    try:
        img_flexj= pil.Image.open(pysedm.io.get_night_files(date, "re:flexuretrace",spec_id, extention=".png")[0])
    except:
        img_flexj= pil.get_buffer([7, 4], "j-flexure image missing" , **prop_missing)
        
    try:
        img_flexi= pil.Image.open(pysedm.io.get_night_files(date, "re:flex_sodium",spec_id, extention=".png")[0])
    except:
        img_flexi= pil.get_buffer([7, 4], "i-flexure image missing" , **prop_missing)

    # Spaxels used:
    try:
        img_spax = pil.Image.open(pysedm.io.get_night_files(date, "re:ifu_spaxels",filesourcename, extention=".png")[0])
    except:
        img_spax = pil.get_buffer([7, 7], "Spaxel IFU image missing" , **prop_missing)

    # ---------
    # PSF Extraction Specials
    # ---------
    # PSF
    try:
        img_psf  = pil.Image.open(pysedm.io.get_night_files(date, "re:psfprofile", filesourcename, extention=".png")[0])
    except:
        img_psf = pil.get_buffer([15, 4], "PSF Profile image missing" , **prop_missing)

    # ADR
    try:
        img_adr  = pil.Image.open(pysedm.io.get_night_files(date, "re:^adr_", filesourcename, extention=".png")[0])
    except:
        img_adr= pil.get_buffer([9, 4], "ADR image missing" , **prop_missing)

    # Output Spectra
    all_spectra_files = pysedm.io.get_night_files(date, "re:^spec_", filesourcename, extention=".png")
    extention = "%s.png" % object_name.split("+")[-1] if "+" in object_name else "%s.png" % object_name
    pysedm_spec_file = pysedm.io.get_night_files(date, "re:^spec_", filesourcename, extention=extention)
    if not STD:
        typed_spectra = [f for f in all_spectra_files if not f.endswith(extention)]
        used_spec_file = pysedm_spec_file if len(typed_spectra) == 0 else typed_spectra
        used_spec_file.sort()
    else:
        calib_spectra = pysedm.io.get_night_files(date, "re:calibcheck", filesourcename, extention=".png")
        used_spec_file = pysedm_spec_file if len(calib_spectra) == 0 else calib_spectra
    try:
        img_spec = pil.Image.open(used_spec_file[-1])
    except:
        img_spec = pil.get_buffer([13, 7], "Spectra image missing", **prop_missing)

    # ============== #
    #  Combination   #
    # ============== #    
    # title
    title = "%s" % object_name
    title += " | exptime: %.1f s" % header["EXPTIME"]
    title += " | Qual: %d" % header["QUALITY"]
    title += " | file ID: %s " % spec_id

    if "aperture" in extraction_mode:
        title_img = pil.get_buffer([7,1], title , fontsize=16, hline=[0.3,0.7], barprop=dict(lw=1))
    
        img_lowerright = pil.get_image_column([img_flexi, img_flexj])
        img_right      = pil.get_image_column([title_img, pil.get_image_row([img_spax,img_lowerright])])
        img_combined   = pil.get_image_row([img_spec, img_right])
        return img_combined

    # Auto

    title_img = pil.get_buffer([8,1], title , fontsize=16, hline=[0.3,0.7], barprop=dict(lw=1))

    # Info 
    width = 3
    flexheader = pil.get_buffer([width,0.8], "Flexure" , fontsize=10, hline=None, textprop=dict(color="0.5"))

    img_upperright = pil.get_image_column([title_img,  img_psf])
    img_upperleft  = pil.get_image_row([img_spax, img_adr])
    img_upper      = pil.get_image_row([img_upperleft, img_upperright])

    bar = pil.get_buffer([0.5,8], vline=0.5, barprop=dict(lw=0.2))
    img_lowerright = pil.get_image_column([flexheader,img_flexi, img_flexj])
    img_lower      = pil.get_image_row([img_spec,bar, img_lowerright])

    img_combined  = pil.get_image_column([img_upper, img_lower, 
                                         footer])
            
    return img_combined


#################################
#
#   MAIN 
#
#################################
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=""" run the interactive plotting of a given cube
            """, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('infile', type=str, default=None,
                        help='cube filepath')

    parser.add_argument('--contains',  type=str, default="*",
                        help='Provide here part of the filename.')

    parser.add_argument('--final',  action="store_true", default=False,
                        help='Push to slack final sedm daily report')

    parser.add_argument('--slack',  action="store_true", default=False,
                        help='Submit the report to slack')
    parser.add_argument('--justpush', action="store_true", default=False,
                        help='Just push to slack')

    # ================ #
    # END of Option    #
    # ================ #
    args = parser.parse_args()

    if args.slack or args.justpush:
        try:
            from pysedmpush import slack
        except ImportError:
            raise ImportError("you need to install pysedmpush"
                              " to be able to push on slack")
        
    # Matplotlib
    # ================= #
    #   The Scripts     #
    # ================= #
    SLACK_CHANNEL = "pysedm-report"
    # --------- #
    #  Date     #
    # --------- #
    date = args.infile

    if args.final:
        print("NOT READY YET")
        

    else:
        # Loads all the spectra 
        specfiles = pysedm.io.get_night_files(date, "spec.basic", args.contains)
        print(specfiles)
        for specfile in specfiles:
            if 'failed' in specfile:
                continue
            report_filename = specfile.replace("spec_","pysedm_report_").replace(".fits",".png")
            if not args.justpush:
                img_report = build_image_report(specfile)
                img_report.save(report_filename, dpi=(1000,1000))

            # Slack push report
            if args.slack or args.justpush:
                if not os.path.isfile( report_filename ):
                    warnings.warn("No file-image created by the pysedm_report.build_image_report(). Nothing to push on slack")
                else:
                    print("pushing the report to %s"%SLACK_CHANNEL)
                    header  = fits.getheader(specfile)
                    file_id = pysedm.io.filename_to_id(specfile)
                    # Title & caption
                    title = "pysedm-report: %s | Qual %d | %s (%s)"%(header["OBJECT"], header["QUALITY"], file_id, date)
                    caption = ""
                    # Push
                    slack.push_image(report_filename, caption=caption, title=title,
                                     channel=SLACK_CHANNEL)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import warnings
from propobject import BaseObject
from pyifu.spectroscopy import Spectrum, get_spectrum
from .sedm import get_sedm_version
""" Library to estimate the flux response based on standard star spectra """

AVOIDANCE_AREA = {"telluric":[[7450,7750],[6850,7050]],
                 "absorption":[[6400,6650],[4810,4910],
                                [4300,4350],[4030,4130],[3900,3980]]}

POLYDEGREE = 40

def get_fluxcalibrator(stdspectrum, polydegree=POLYDEGREE, fullout=False, filter=None):
    """ fit a smooth polynome through the ratio between observed standard and their expected flux (calspec) """
    fl = FluxCalibrator()
    fl.set_std_spectrum(stdspectrum)
    return  fl.fit_inverse_sensitivity(polydegree=polydegree, filter=filter), fl

def show_fluxcalibrated_standard(stdspectrum, savefile=None):
    """ """
    import matplotlib.pyplot as mpl
    import pycalspec
    from astropy import units
    from . import io
    ### Data
    objectname = stdspectrum.header['OBJECT'].replace("STD-","").split()[0]
    #
    try:
        specref = pycalspec.std_spectrum(objectname).filter(5).reshape(stdspectrum.lbda,"linear")
    except IOError:
        print("WARNING: ref flux not found so no flux scale plot possible")
        return None
    specres = get_spectrum(stdspectrum.lbda, specref.data / stdspectrum.data )
    scale_ratio = specres.data.mean()
    specres.scale_by(scale_ratio)
    #
    try:
        dtime = io.filename_to_time( stdspectrum.filename) - io.filename_to_time( stdspectrum.header["CALSRC"] )
    except:
        dtime = None
    ###

    fig = mpl.figure(figsize=[6,4])

    ax   = fig.add_axes([0.13,0.43,0.8,0.5])
    axr  = fig.add_axes([0.13,0.13,0.8,0.28])

    stdspectrum.show(ax=ax, label="observation", show=False, yscalefill=True)
    specref.show(ax=ax, color="C1", label="calspec", show=False, yscalefill=True)

    axr.axhline(1, ls="--", color="0.5")
    axr.axhspan(0.9,1.1, color="0.5", alpha=0.1)
    axr.axhspan(0.8,1.2, color="0.5", alpha=0.1)
    # Residual
    specres.show(ax=axr, color="k", lw=1.5, show=False)

    ax.set_yscale("log")
    for l in AVOIDANCE_AREA["telluric"]: 
        ax.axvspan(l[0],l[1], color=mpl.cm.Blues(0.6), alpha=0.2)
        axr.axvspan(l[0],l[1], color=mpl.cm.Blues(0.6), alpha=0.2)
        
    axr.text(0.02,0.92, "scale ratio: %.1f"%scale_ratio, transform=axr.transAxes,
                va="top", ha="left", backgroundcolor= mpl.cm.binary(0.,0.8))

    ax.set_xlabel(["" for l in ax.get_xlabel()])
    ax.set_ylabel(r"Flux [erg s$^{-1}$ cm$^{-2}$ A$^{-1}$]")
    if dtime is not None:
        ax.set_title( "%s"%objectname+ r" | t$_{obs}$ - t$_{fluxcal}$ : %.2f hours"%((dtime.sec * units.second).to("hour")).value)
    else:
        ax.set_title( "%s"%objectname)

    axr.set_ylabel(r"Flux ratio")
    axr.set_xlabel(r"Wavelength [$\AA$]")

    ax.legend(loc="lower left")
    if savefile:
        fig.savefig(savefile)
    else:
        return fig


# =================== #
#  INTERNAL Tools     #
# =================== #
_TELL_HEADERKEY = "TELL"
def _paramkey_to_headerkey_( paramkey):
    """ converts the argument names from TelluricLines into Fits format header entry """
    return _TELL_HEADERKEY+paramkey.replace("coef","c").replace("rho_","r").replace("filter","fltr").replace("amplitude","AMP").upper()
def _headerkey_to_paramkey_( headerkey ):
    """ converts  Fits format header entry into argument names from TelluricLines into """
    return  headerkey.replace(_TELL_HEADERKEY+"C","coef").replace(_TELL_HEADERKEY+"R","rho_").replace(_TELL_HEADERKEY+"FLTR","filter").replace(_TELL_HEADERKEY+"AMP","amplitude").lower()

    
def get_fluxcal_spectrum(lbda, flux, variance=None, header=None, logwave=None):
    """ Create a Flux Calibrator spectrum from the given data
    
    Parameters
    ----------
    lbda: [array]
        wavelength of the spectrum

    flux: [array]
        flux of the spectrum
    
    variance: [array/None] -optional-
        variance associated to the flux.

    header: [fits header / None]
        fits header assoiated to the data
        
    logwave: [None / bool] -optional-
        If the wavelength given in log of wavelength. 
        If you known set True (= given in log) or False (= given in angstrom)
        If let to None, this will test if the first wavelength is smaller or 
        higher than a default number (50).


    Returns
    -------
    Spectrum
    """
    spec = FluxCalSpectrum(None)
    spec.create(data=flux, variance=variance, header=header, lbda=lbda, logwave=logwave)
    return spec


def load_fluxcal_spectrum(filename,**kwargs):
    """ """
    return FluxCalSpectrum(filename, **kwargs)

class FluxCalSpectrum( Spectrum ):
    """ """
    PROPERTIES      = ["tellspec","tellparam"]
    SIDE_PROPERTIES = ["refairmass", "ref_amplitude"]

    # ================= #
    #   Methods         #
    # ================= #
    # -------- #
    #  SETTER  #
    # -------- #
    def set_header(self, header):
        """ 
        Attach a header. 
        If the given header is None, an empty header will be attached.

        If Telluric parameter information are found in the header. They are set.
        """
        super(FluxCalSpectrum, self).set_header(header)
        [self.set_telluric_parameters(**{_headerkey_to_paramkey_(k):v})
             for k,v in self.header.items() if k.startswith(_TELL_HEADERKEY)]
        
    def set_telluric_parameters(self, **kwargs):
        """ Telluric parameter could be:
        - coefo2, 
        - coefh2o, 
        - rho_o2, 
        - rho_h2o, 
        - filter 
        - amplitude """
        VALIDS =["coefo2", "coefh2o",
                 "rho_o2", "rho_h2o",
                "filter","amplitude"]
        
        for k,v in kwargs.items():
            if k not in VALIDS:
                raise ValueError("%s is not a valid telluric parameter"%k +"these are: "+ ", ".join(VALIDS))
            if k in ["amplitude"]:
                self._side_properties["ref_amplitude"] = v
            else:
                self.tellparam[k] = v

    def load_telluric_spectrum(self):
        """ """
        from .io import load_telluric_line
        self._properties['tellspec'] = load_telluric_line()
        
    # -------- #
    #  GETTER  #
    # -------- #
    def get_inversed_sensitivity(self, airmass, amplitude=None, **kwargs):
        """ """
        if self.is_old_format:
            return 1/self.data
        
        if amplitude is None: amplitude = self.ref_amplitude
        self.set_telluric_parameters(**kwargs)
        return self.data + self.get_telluric_throughput(airmass, amplitude=amplitude)
    
    def get_telluric_absorption(self, airmass, amplitude=None,**kwargs):
        """ """
        if self.is_old_format:
            return 0
        if amplitude is None: amplitude = self.ref_amplitude
        self.set_telluric_parameters(**kwargs)
        return amplitude*self.tellspec.get_telluric_absorption(airmass, **self.tellparam).reshape(self.lbda, "linear").data

    def get_telluric_throughput(self, airmass, amplitude=None, **kwargs):
        """ """
        if self.is_old_format:
            return 0
        if amplitude is None: amplitude = self.ref_amplitude
        self.set_telluric_parameters(**kwargs)
        return amplitude*self.tellspec.get_telluric_throughput(airmass, **self.tellparam).reshape(self.lbda, "linear").data

    
    # -------- #
    # PLOTTER  #
    # -------- #
    def show(self, ax=None, airmass=1, color="C0", fillcolor="0.7", fillalpha=0.7,
                 inversed=False, show=True, zorder=4, telllabel="_no_legend_", label=None,**kwargs):
        """ """
        import matplotlib.pyplot as mpl
        if ax is None:
            fig = mpl.figure(figsize=[7,4])
            ax  = fig.add_axes([0.12,0.15,0.78,0.75])
        else:
            fig = ax.figure
        
        power= 1 if not inversed else -1
        if not self.is_old_format:
            ax.fill_between(self.lbda, self.data**power, self.get_inversed_sensitivity(airmass)**power, 
                            color=fillcolor, alpha=0.7,zorder=zorder-1, label=telllabel)
        ax.plot( self.lbda, self.get_inversed_sensitivity(airmass)**power, color=color, zorder=zorder,
                     label=label,**kwargs)
        if show:
            fig.show()
            
        return fig
        
            
    # ================= #
    #   Properties      #
    # ================= #
    def has_tellspec(self):
        """ """
        return self._properties['tellspec'] is not None
    
    @property
    def tellparam(self):
        """ """
        if self._properties["tellparam"] is None:
            self._properties["tellparam"] = {}
        return self._properties["tellparam"]
    
    @property
    def tellspec(self):
        """ """
        if not self.has_tellspec(): self.load_telluric_spectrum()
        return self._properties['tellspec']

    @property
    def ref_amplitude(self):
        """ amplitude of the telluric line in during the calibration """
        return self._side_properties["ref_amplitude"]

    @property
    def is_old_format(self):
        """ """
        return "TELLFLTR" not in self.header

    
class FluxCalibrator( BaseObject ):
    """ """
    PROPERTIES = ["spectrum", "calspec"]
    DERIVED_PROPERTIES = ['calspec_spectrum', "fluxcalspectrum"]
    # ==================== #
    #  Method              #
    # ==================== #
    # -------- #
    # GETTER   #
    # -------- #
    def fit_inverse_sensitivity(self, polydegree=POLYDEGREE, filter=4, masked_area=None,**kwargs):
        """ **kwargs goes to polyfit.fit(**kwargs)
        
        Parameters
        ----------
        filter: [None, float] -optional-
            if not None, the free_parameter 'filter' will be forced to the given value (hence not fitted)
                This is similar as doing: `filter_guess=XX, filter_fixed=True`
            if None, nothing will be made.
            If you only want to give initial guess and e.g. boundaries:
               - set filter to None: filter=None
               - provide the fit entry: filter_guess=XX, filter_boundaries=[YY,ZZ]
               
        """
        from .utils.atmosphere import TelluricPolynomeFit
        from .io import load_telluric_line
        from modefit import get_polyfit

        if masked_area is not None:
            maskin = self.get_avoidance_mask("absorption")
        else:
            maskin = None
            
        datacal = self.spectrum.data/self.calspec_spectrum.data
        norm = datacal.mean()
        datacal/=norm

        # Based on least sq
        #errcal = datacal * (np.sqrt(self.spectrum.variance)/self.spectrum.data)# + np.sqrt(self.calspec_spectrum.variance)/self.calspec_spectrum.data)
        errcal = datacal/100
        
        
        if filter is not None:
            kwargs["filter_guess"] = filter
            kwargs["filter_fixed"] = True
        else:
            kwargs["filter_guess"] = 18

        
        end_airmass = float(self.spectrum.header["ENDAIR"]) if "ENDAIR" in self.spectrum.header else self.spectrum.header["AIRMASS"]
        start_airmass = float(self.spectrum.header["AIRMASS"])
        airmass = np.mean([start_airmass, end_airmass])
        ##
        # Older procedure
        ##
        ##
        # New procedure
        ##
        # A: 3 Step procedure.
        # - Step 1 fit for the telluric region
        tell_range = (self.spectrum.lbda>6000) & (self.spectrum.lbda<8500)
        POLY_TELL  = 5
        self.tpoly_tell = TelluricPolynomeFit(self.spectrum.lbda[tell_range],
                                              datacal[tell_range],
                                              errcal[tell_range],
                                              POLY_TELL, load_telluric_line(),
                                                  maskin=None)
        self.tpoly_tell.fit(airmass_guess= airmass, airmass_fixed= True, airmass_boundaries=np.sort([start_airmass, end_airmass])*[1, 1.05],
                      filter_guess=15, filter_fixed=False)
        
        # use this to fix telluric info in for the second run
        kwarg_fit = {**{k+"_guess":self.tpoly_tell.fitvalues[k] for k in self.tpoly_tell.model.TELL_FREEPARAMETERS},
                         **{k+"_fixed":True for k in self.tpoly_tell.model.TELL_FREEPARAMETERS}}
        
        # - Step 2 fit The blue region
        if get_sedm_version(self.spectrum.header.get("OBSDATE",None)):
            warnings.warn("SEDM version <3 (pre-Feb 2019): running specific interpolation procedure.")
            POLY_TELL_BLUE = 20
            self.tpoly_blue = TelluricPolynomeFit(self.spectrum.lbda,
                                                  datacal,
                                                  errcal,
                                                  POLY_TELL_BLUE, load_telluric_line(), maskin=maskin)
            self.tpoly_blue.fit(**kwarg_fit)
            self.tpoly_blue.norm = norm
            response_continuum = self.tpoly_blue.model._get_continuum_()
            POLY_TELL = "%d"%POLY_TELL_BLUE
        else:
            blue_range = (self.spectrum.lbda<8600)
            POLY_TELL_BLUE = 20
            self.tpoly_blue = TelluricPolynomeFit(self.spectrum.lbda[blue_range],
                                                  datacal[blue_range],
                                                  errcal[blue_range],
                                                  POLY_TELL_BLUE, load_telluric_line(), maskin=maskin)
            self.tpoly_blue.fit(**kwarg_fit)
            self.tpoly_blue.norm = norm
            #
            # - Step 3 fit The red region
            red_range = (self.spectrum.lbda>8400)
            POLY_TELL_RED = 15
            telluric = self.tpoly_blue.model.get_telluric_model(lbda=self.spectrum.lbda)
        
            self.poly_red = get_polyfit(self.spectrum.lbda[red_range], 
                                        (datacal-telluric)[red_range], 
                                        errcal[red_range], POLY_TELL_RED, legendre=True)
            self.poly_red.fit()
            #
            # Finally Merge the two continuums
            #
            cont_blue = np.ones(len(self.spectrum.lbda))*np.NaN
            cont_blue[blue_range] = self.tpoly_blue.model._get_continuum_()
        
            cont_red  = np.ones(len(self.spectrum.lbda))*np.NaN
            cont_red[red_range] = self.poly_red.model.get_model()
        
            response_continuum = np.nanmean([cont_blue, cont_red], axis=0)
            POLY_TELL = "%d,%d"%(POLY_TELL_BLUE,POLY_TELL_RED)
        #
        # END
        #

        
        header = self.spectrum.header.copy()
    
        self.tpoly_blue.fitvalues['amplitude'] *= self.tpoly_blue.norm
        for tellk in self.tpoly_blue.model.TELL_FREEPARAMETERS:
            if tellk in ["airmass"]: continue
            header.set(_paramkey_to_headerkey_(tellk), self.tpoly_blue.fitvalues[tellk], "Telluric Parameter")
            
        header.set("CONTDEG", POLY_TELL,  "degree of the continuum polynome (legendre)")
        self._derived_properties["fluxcalspectrum"] = get_fluxcal_spectrum(self.spectrum.lbda, response_continuum*self.tpoly_blue.norm, header=header)
        
        return self.fluxcalspectrum

    def get_avoidance_mask(self, which="both"):
        """ boolean mask.
        Parameters
        ----------
        which: [string] -optional-
            - both: False if wavelength in any Telluric or Absorption intervals
            - telluric: False if wavelength in Telluric intervals
            - absorption: False if wavelength in Absorption intervals
        """
        flagin_tell = np.any([(self.spectrum.lbda>l[0]) * (self.spectrum.lbda<l[1]) for l in AVOIDANCE_AREA["telluric"]], axis=0)
        flagin_abs = np.any([(self.spectrum.lbda>l[0]) * (self.spectrum.lbda<l[1]) for l in AVOIDANCE_AREA["absorption"]], axis=0)
        if which in ["both","Both"]:
            return ~(flagin_tell+flagin_abs)
        if which in ["telluric","Telluric", "tell"]:
            return ~flagin_tell
        if which in ["absorption","Absorption", "Abs"]:
            return ~flagin_abs
        raise ValueError("Unknown masking request %s (use which/telluric/absorption)"%which)

    # -------- #
    #  SETTER  #
    # -------- #
    def set_std_spectrum(self, spectrum, load_reference=True):
        """ """
        self._properties['spectrum'] = spectrum
        if load_reference:
            self.load_calspec_reference()

    def load_calspec_reference(self):
        """ """
        try:
            import pycalspec
        except ImportError:
            raise ImportError("You need pycalspec. Please pip install pycalspec (or grab it on github)")
        
        self._properties['calspec'] = pycalspec.std_spectrum(self.objectname)
        self._derived_properties['calspec_spectrum'] = self._calspec.filter(5).reshape(self.spectrum.lbda,"linear")

    # --------- #
    #  PLOTTER  #
    # --------- #
    def show(self, savefile=None, show=True, ratiocolor="k", fluxcalcolor="C2",
                 show_vertbands=True, **kwargs):
        """ """
        import matplotlib.pyplot as mpl
        from pyifu import get_spectrum
                
        fig = mpl.figure()
        ax   = fig.add_axes([0.15,0.64,0.8,0.3])
        ax2  = fig.add_axes([0.15,0.12,0.8,0.5])
        axt  = ax.twinx()
        self.spectrum.show(ax=ax, show=False)
        self.calspec_spectrum.show(ax=axt, color="C1", show=False)
        
        for avoidance,c in zip(AVOIDANCE_AREA.values(), [mpl.cm.Blues(0.6),"0.5"]) :
            for l in avoidance: ax.axvspan(l[0],l[1], color=c, alpha=0.2)
        
        spec = get_spectrum(self.spectrum.lbda, self.calspec_spectrum.data / self.spectrum.data )

        spec.show(ax=ax2, color=ratiocolor, label="Calspec / SEDM", show=False, zorder=5, **kwargs)
        if self.fluxcalspectrum is not None:
            self.fluxcalspectrum.show(ax=ax2, color=fluxcalcolor, label="Flux Calibration",
                                          telllabel="incl. telluric", show=False, inversed=True, zorder=4)
            
        ax2.set_yscale("log")

        ax.set_xticks([])
        ax.set_yticks([])
        axt.set_yticks([])
        ax2.set_xlabel(r"Wavelength [$\AA$]", fontsize="large")
        ax2.set_ylabel("Inversed Sensitivity", fontsize="large")
        ax2.legend(loc="best")
        if savefile is not None:
            fig.savefig(savefile)
        if show:
            fig.show()
            
    # ==================== #
    #  Properties          #
    # ==================== #
    @property
    def spectrum(self):
        """ Standard star spectrum """
        return self._properties['spectrum']
    
    @property
    def objectname(self):
        """ """
        return self.spectrum.header['OBJECT'].replace("STD-","").split()[0]

    # - CalSpec
    @property
    def _calspec(self):
        """ The CalSpec Spectrum as given by calspec"""
        return self._properties['calspec']
    
    @property
    def calspec_spectrum(self):
        """ CalSpec Spectrum at the `spectrum` wavelength interpolation """
        return self._derived_properties['calspec_spectrum']
    
    @property
    def fluxcalspectrum(self):
        """ """
        return self._derived_properties["fluxcalspectrum"]
